//
//  ABFilterAudioEngine.m
//  Audiobus Samples
//
//  Created by Michael Tyson on 18/12/2013.
//  Copyright (c) 2011-2014 Audiobus. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without modification,
//  are permitted provided that the following conditions are met:
//
//  - Redistributions of source code must retain the above copyright notice, this
//    list of conditions and the following disclaimer.
//
//  - Redistributions in binary form must reproduce the above copyright notice, this
//    list of conditions and the following disclaimer in the documentation and/or
//    other materials provided with the distribution.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
//  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
//  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
//  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#import "ABFilterAudioEngine.h"

#import "ABFilterAudioEngine.h"
#import <libkern/OSAtomic.h>
#import <AVFoundation/AVFoundation.h>
#import "Audiobus.h"

#define checkResult(result,operation) (_checkResult((result),(operation),strrchr(__FILE__, '/')+1,__LINE__))
static inline BOOL _checkResult(OSStatus result, const char *operation, const char* file, int line) {
    if ( result != noErr ) {
        NSLog(@"%s:%d: %s result %d %08X %4.4s\n", file, line, operation, (int)result, (int)result, (char*)&result);
        return NO;
    }
    return YES;
}

static void * kMemberOfActiveAudiobusSessionChanged = &kMemberOfActiveAudiobusSessionChanged;
static void * kAudiobusConnectedChanged = &kAudiobusConnectedChanged;

@interface ABFilterAudioEngine () {
    AudioUnit _ioUnit;
    AudioUnit _delayUnit;
}
@property (strong, nonatomic, readwrite) ABAudiobusController *audiobusController;
@property (strong, nonatomic) ABFilterPort *filter;
@property (nonatomic, strong) NSArray *observerTokens;
@property (nonatomic, assign) BOOL muted;
@end

@implementation ABFilterAudioEngine
@dynamic wetDryMix, delayTime, feedback, lopassCutoff;

-(id)init {
    if ( !(self = [super init]) ) return nil;
    
    [self setup];
    
    // Create an Audiobus instance
    self.audiobusController = [[ABAudiobusController alloc] initWithApiKey:@"MCoqKkFCIEZpbHRlcioqKmFiZmlsdGVyLmF1ZGlvYnVzOi8vKioqW2F1cnguYWJ1cy5kZWxhLjFd:VX+6+AsiaLneF6zEpaDsQIIrYoIuF5bpsnCJg8DGVC4G1/0Rl5MyLtiIGbS3cWR+wtXi4LwRMbfZfjs4Z3TPzM9QgTiZoDAOtDKPQVf59sWt6YZ5z7jbTOyqLUKvN6tk"];
    
    // Create a filter port
    self.filter = [[ABFilterPort alloc] initWithName:@"Delay"
                                               title:@"Delay"
                           audioComponentDescription:(AudioComponentDescription) {
                               .componentType = kAudioUnitType_RemoteEffect,
                               .componentSubType = 'dela',
                               .componentManufacturer = 'abus' }
                                           audioUnit:_ioUnit];
    
    [_audiobusController addFilterPort:_filter];
    
    // Watch the connected and memberOfActiveAudiobusSession properties to be notified when we connect/disconnect or the Audiobus session begins/ends
    [_audiobusController addObserver:self forKeyPath:@"connected" options:0 context:kAudiobusConnectedChanged];
    [_audiobusController addObserver:self forKeyPath:@"memberOfActiveAudiobusSession" options:0 context:kMemberOfActiveAudiobusSessionChanged];

    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(applicationDidFinishLaunching:) name:UIApplicationDidFinishLaunchingNotification object:nil];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(applicationDidEnterBackground:) name:UIApplicationDidEnterBackgroundNotification object:nil];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(applicationWillEnterForeground:) name:UIApplicationWillEnterForegroundNotification object:nil];
    
    return self;
}

-(void)dealloc {
    [_audiobusController removeObserver:self forKeyPath:@"connected"];
    [_audiobusController removeObserver:self forKeyPath:@"memberOfActiveAudiobusSession"];
    [self teardown];
    [[NSNotificationCenter defaultCenter] removeObserver:self];
}

-(void)setWetDryMix:(Float32)wetDryMix {
    AudioUnitParameterValue value = wetDryMix;
    checkResult(AudioUnitSetParameter(_delayUnit, kDelayParam_WetDryMix, kAudioUnitScope_Global, 0, value, 0),
                "AudioUnitSetParameter(kDelayParam_WetDryMix)");
}

-(Float32)wetDryMix {
    AudioUnitParameterValue value = 0.0;
    checkResult(AudioUnitGetParameter(_delayUnit, kDelayParam_WetDryMix, kAudioUnitScope_Global, 0, &value),
                "AudioUnitGetParameter(kDelayParam_WetDryMix)");
    return value;
}

-(void)setDelayTime:(Float32)delayTime {
    AudioUnitParameterValue value = delayTime;
    checkResult(AudioUnitSetParameter(_delayUnit, kDelayParam_DelayTime, kAudioUnitScope_Global, 0, value, 0),
                "AudioUnitSetParameter(kDelayParam_DelayTime)");
}

-(Float32)delayTime {
    AudioUnitParameterValue value = 0.0;
    checkResult(AudioUnitGetParameter(_delayUnit, kDelayParam_DelayTime, kAudioUnitScope_Global, 0, &value),
                "AudioUnitGetParameter(kDelayParam_DelayTime)");
    return value;
}

-(void)setFeedback:(Float32)feedback {
    AudioUnitParameterValue value = feedback;
    checkResult(AudioUnitSetParameter(_delayUnit, kDelayParam_Feedback, kAudioUnitScope_Global, 0, value, 0),
                "AudioUnitSetParameter(kDelayParam_Feedback)");
}

-(Float32)feedback {
    AudioUnitParameterValue value = 0.0;
    checkResult(AudioUnitGetParameter(_delayUnit, kDelayParam_Feedback, kAudioUnitScope_Global, 0, &value),
                "AudioUnitGetParameter(kDelayParam_Feedback)");
    return value;
}

-(void)setLopassCutoff:(Float32)lopassCutoff {
    AudioUnitParameterValue value = lopassCutoff;
    checkResult(AudioUnitSetParameter(_delayUnit, kDelayParam_LopassCutoff, kAudioUnitScope_Global, 0, value, 0),
                "AudioUnitSetParameter(kDelayParam_LopassCutoff)");
}

-(Float32)lopassCutoff {
    AudioUnitParameterValue value = 0.0;
    checkResult(AudioUnitGetParameter(_delayUnit, kDelayParam_LopassCutoff, kAudioUnitScope_Global, 0, &value),
                "AudioUnitGetParameter(kDelayParam_LopassCutoff)");
    return value;
}

-(void)setMuted:(BOOL)muted {
    if ( _muted == muted ) return;
    _muted = muted;
    
    if ( _muted ) {
        // Break connection to output
        checkResult(AudioUnitSetProperty(_ioUnit, kAudioUnitProperty_MakeConnection, kAudioUnitScope_Input, 0, &(AudioUnitConnection) {
            .sourceAudioUnit = NULL,
            .sourceOutputNumber = 0,
            .destInputNumber = 0
        }, sizeof(AudioUnitConnection)), "AudioUnitSetProperty(kAudioUnitProperty_MakeConnection)");
    } else {
        // Re-make connection to output
        checkResult(AudioUnitSetProperty(_ioUnit, kAudioUnitProperty_MakeConnection, kAudioUnitScope_Input, 0, &(AudioUnitConnection) {
            .sourceAudioUnit = _delayUnit,
            .sourceOutputNumber = 0,
            .destInputNumber = 0
        }, sizeof(AudioUnitConnection)), "AudioUnitSetProperty(kAudioUnitProperty_MakeConnection)");
    }
}

- (void)setup {
    
    NSError *error = nil;
    if ( ![[AVAudioSession sharedInstance] setCategory:AVAudioSessionCategoryPlayAndRecord
                                           withOptions:AVAudioSessionCategoryOptionMixWithOthers | AVAudioSessionCategoryOptionDefaultToSpeaker
                                                 error:&error] ) {
        NSLog(@"Couldn't set audio session category: %@", error);
    }
    
    if ( ![[AVAudioSession sharedInstance] setActive:YES error:&error] ) {
        NSLog(@"Couldn't set audio session active: %@", error);
    }
    
    // Create the audio unit
    AudioComponentDescription desc = {
        .componentType = kAudioUnitType_Output,
        .componentSubType = kAudioUnitSubType_RemoteIO,
        .componentManufacturer = kAudioUnitManufacturer_Apple,
        .componentFlags = 0,
        .componentFlagsMask = 0
    };
    
    AudioComponent component = AudioComponentFindNext(NULL, &desc);
    checkResult(AudioComponentInstanceNew(component, &_ioUnit), "AudioComponentInstanceNew");
    
    // Enable rendering while screen locked
    UInt32 framesPerSlice = 4096;
    checkResult(AudioUnitSetProperty(_ioUnit, kAudioUnitProperty_MaximumFramesPerSlice, kAudioUnitScope_Global, 0, &framesPerSlice, sizeof(framesPerSlice)),
                "AudioUnitSetProperty(kAudioUnitProperty_MaximumFramesPerSlice");
    
    // Enable input
    UInt32 enableFlag = 1;
    checkResult(AudioUnitSetProperty(_ioUnit, kAudioOutputUnitProperty_EnableIO, kAudioUnitScope_Input, 1, &enableFlag, sizeof(enableFlag)),
                "kAudioOutputUnitProperty_EnableIO");
    
    // Create and hook up the delay unit
    desc = (AudioComponentDescription){
        .componentType = kAudioUnitType_Effect,
        .componentSubType = kAudioUnitSubType_Delay,
        .componentManufacturer = kAudioUnitManufacturer_Apple,
        .componentFlags = 0,
        .componentFlagsMask = 0
    };
    
    component = AudioComponentFindNext(NULL, &desc);
    checkResult(AudioComponentInstanceNew(component, &_delayUnit), "AudioComponentInstanceNew");
    
    // Enable rendering while screen locked
    checkResult(AudioUnitSetProperty(_delayUnit, kAudioUnitProperty_MaximumFramesPerSlice, kAudioUnitScope_Global, 0, &framesPerSlice, sizeof(framesPerSlice)),
                "AudioUnitSetProperty(kAudioUnitProperty_MaximumFramesPerSlice");
    
    checkResult(AudioUnitSetProperty(_delayUnit, kAudioUnitProperty_MakeConnection, kAudioUnitScope_Input, 0, &(AudioUnitConnection) {
        .sourceAudioUnit = _ioUnit,
        .sourceOutputNumber = 1,
        .destInputNumber = 0
    }, sizeof(AudioUnitConnection)), "AudioUnitSetProperty(kAudioUnitProperty_MakeConnection)");
    
    checkResult(AudioUnitSetProperty(_ioUnit, kAudioUnitProperty_MakeConnection, kAudioUnitScope_Input, 0, &(AudioUnitConnection) {
        .sourceAudioUnit = _delayUnit,
        .sourceOutputNumber = 0,
        .destInputNumber = 0
    }, sizeof(AudioUnitConnection)), "AudioUnitSetProperty(kAudioUnitProperty_MakeConnection)");
    
    [self updateChannelMap];
    
    // Initialize the audio units
    checkResult(AudioUnitInitialize(_ioUnit), "AudioUnitInitialize");
    checkResult(AudioUnitInitialize(_delayUnit), "AudioUnitInitialize");
    
    // Watch for session interruptions
    self.observerTokens = @[];
    self.observerTokens = [_observerTokens arrayByAddingObject:
        [[NSNotificationCenter defaultCenter] addObserverForName:AVAudioSessionInterruptionNotification object:nil queue:nil usingBlock:^(NSNotification *notification) {
            NSInteger type = [notification.userInfo[AVAudioSessionInterruptionTypeKey] integerValue];
            if ( type == AVAudioSessionInterruptionTypeBegan ) {
                [self stop];
            } else {
                if ( ![self start] ) {
                    // Work around an iOS 7 audio interruption bug
                    [self teardown];
                    [self setup];
                    [self start];
                }
            }
        }]];
    
    // Watch for route change notifications
    self.observerTokens = [_observerTokens arrayByAddingObject:
        [[NSNotificationCenter defaultCenter] addObserverForName:AVAudioSessionRouteChangeNotification object:nil queue:nil usingBlock:^(NSNotification *notification) {
            [self updateChannelMap];
        }]];
}

- (void)teardown {
    if ( self.running ) {
        [self stop];
    }
    if ( _ioUnit ) {
        checkResult(AudioComponentInstanceDispose(_ioUnit), "AudioComponentInstanceDispose");
        _ioUnit = NULL;
    }
    if ( _delayUnit ) {
        checkResult(AudioComponentInstanceDispose(_delayUnit), "AudioComponentInstanceDispose");
        _delayUnit = NULL;
    }
    for ( id token in _observerTokens ) {
        [[NSNotificationCenter defaultCenter] removeObserver:token];
    }
    self.observerTokens = nil;
}

- (BOOL)running {
    Boolean running = false;
    UInt32 size = sizeof(running);
    if ( !checkResult(AudioUnitGetProperty(_ioUnit, kAudioOutputUnitProperty_IsRunning, kAudioUnitScope_Global, 0, &running, &size), "AudioUnitGetProperty") ) {
        return NO;
    }
    
    return running;
}

- (void)stop {
    NSLog(@"Stopping audio engine");
    if ( _ioUnit ) {
        checkResult(AudioOutputUnitStop(_ioUnit), "AudioOutputUnitStop");
        [[AVAudioSession sharedInstance] setActive:NO error:NULL];
    }
}

- (BOOL)start {
    NSLog(@"Starting audio engine");
    NSError *error = nil;
    if ( ![[AVAudioSession sharedInstance] setActive:YES error:&error] ) {
        NSLog(@"Couldn't activate audio session: %@", error);
        return NO;
    }
    return checkResult(AudioOutputUnitStart(_ioUnit), "AudioOutputUnitStart");
}

-(void)updateChannelMap {
    // Update channel map to ensure if we're stereo out but mono in, we hear the input in both output channels
    NSInteger numberInputChannels = [[AVAudioSession sharedInstance] inputNumberOfChannels];
    UInt32 channelMap[2] = { 0, numberInputChannels > 1 ? 1 : 0 };
    checkResult(AudioUnitSetProperty(_ioUnit, kAudioOutputUnitProperty_ChannelMap, kAudioUnitScope_Output, 1, channelMap, sizeof(channelMap)), "kAudioOutputUnitProperty_ChannelMap");
}

#pragma mark - Events

-(void)applicationDidFinishLaunching:(NSNotification*)notification {
    if ( [[notification.userInfo[UIApplicationLaunchOptionsURLKey] scheme] hasSuffix:@".audiobus"] ) {
        // If this effect app has been launched from within Audiobus, we need to silence our output for a little while
        // to avoid feedback issues while the connection is established.
        if ( !_audiobusController.connected ) {
            // Mute
            self.muted = YES;
            
            // Set a timeout for three seconds, after which we can assume there's actually no
            // Audiobus connection forthcoming (something went wrong), and we should unmute
            dispatch_after(dispatch_time(DISPATCH_TIME_NOW, (int64_t)(3.0 * NSEC_PER_SEC)), dispatch_get_main_queue(), ^{
                self.muted = NO;
            });
        }
    }
}

-(void)applicationDidEnterBackground:(NSNotification *)notification {
    if ( !_audiobusController.connected && !_audiobusController.memberOfActiveAudiobusSession ) {
        // Fade out and stop the audio engine, suspending the app, if we're not connected, and we're not part of an active Audiobus session
        [ABAudioUnitFader fadeOutAudioUnit:_ioUnit completionBlock:^{ [self stop]; }];
    }
}

-(void)applicationWillEnterForeground:(NSNotification *)notification {
    if ( !self.running || [ABAudioUnitFader transitionsRunning] ) {
        // Start the audio system and fade in if it wasn't running
        [ABAudioUnitFader fadeInAudioUnit:_ioUnit beginBlock:^{ [self start]; } completionBlock:nil];
    }
}

-(void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context {
    if ( context == kMemberOfActiveAudiobusSessionChanged || context == kAudiobusConnectedChanged ) {
        if ( [UIApplication sharedApplication].applicationState == UIApplicationStateBackground
                && !_audiobusController.connected
                && !_audiobusController.memberOfActiveAudiobusSession ) {
            // Audiobus session is finished. Time to sleep.
            [self stop];
        } else if ( context == kAudiobusConnectedChanged && _audiobusController.connected ) {
            
            // We just connected to Audiobus (or IAA). Reset delay unit, to clear the delay lines
            checkResult(AudioUnitReset(_delayUnit, kAudioUnitScope_Global, 0), "AudioUnitReset");
            
            // Make sure we're running
            [self start];
            
            if ( _muted ) {
                // Unmute now if we were launched from Audiobus
                self.muted = NO;
            }
        }
    } else {
        [super observeValueForKeyPath:keyPath ofObject:object change:change context:context];
    }
}

@end
