//
//  ABMultitrackAudioEngine.m
//  Audiobus Samples
//
//  Created by Michael Tyson on 18/12/2013.
//  Copyright (c) 2011-2014 Audiobus. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without modification,
//  are permitted provided that the following conditions are met:
//
//  - Redistributions of source code must retain the above copyright notice, this
//    list of conditions and the following disclaimer.
//
//  - Redistributions in binary form must reproduce the above copyright notice, this
//    list of conditions and the following disclaimer in the documentation and/or
//    other materials provided with the distribution.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
//  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
//  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
//  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#import "ABMultitrackAudioEngine.h"
#import <libkern/OSAtomic.h>
#import "Audiobus.h"
#import <AVFoundation/AVFoundation.h>
#import <Accelerate/Accelerate.h>

#define checkResult(result,operation) (_checkResult((result),(operation),strrchr(__FILE__, '/')+1,__LINE__))
static inline BOOL _checkResult(OSStatus result, const char *operation, const char* file, int line) {
    if ( result != noErr ) {
        NSLog(@"%s:%d: %s result %d %08X %4.4s\n", file, line, operation, (int)result, (int)result, (char*)&result);
        return NO;
    }
    return YES;
}

NSString * const ABMultitrackAudioEnginePortAddedNotification = @"ABMultitrackAudioEnginePortAddedNotification";
NSString * const ABMultitrackAudioEnginePortRemovedNotification = @"ABMultitrackAudioEnginePortRemovedNotification";

NSString * const ABMultitrackAudioEnginePortKey = @"port";

static void * kAudiobusConnectedOrActiveMemberChanged = &kAudiobusConnectedOrActiveMemberChanged;

static const int kMaxReceiveBlocks = 10;
static const int kMaxSources = 10;

struct port_entry_t {
    void *portPtr; // We use void*s as ARC forbids Objective-C objects in structs, but we don't want to use any Objective-C on the realtime thread
    void *blockPtrs[kMaxReceiveBlocks];
    BOOL pendingRemoval;
};

@interface ABMultitrackAudioEngine () {
    AudioUnit _ioUnit;
    AudioUnit _mixerUnit;
    struct port_entry_t _ports[kMaxSources];
}
@property (strong, nonatomic, readwrite) ABAudiobusController *audiobusController;
@property (strong, nonatomic) ABReceiverPort *input;
@property (nonatomic, strong) NSMutableDictionary *receiveBlockDictionary;
@property (nonatomic, strong) id observerToken;
@end

@implementation ABMultitrackAudioEngine
@dynamic connectedPorts;

-(id)init {
    if ( !(self = [super init]) ) return nil;
    
    [self setup];
    
    // Create an Audiobus instance
    self.audiobusController = [[ABAudiobusController alloc] initWithApiKey:@"MCoqKkFCIE11bHRpdHJhY2sqKiphYm11bHRpdHJhY2suYXVkaW9idXM6Ly8qKio=:M1B/GSVKOhe1Kep5kzwoq/jgCxNQEuec3o/4TK8URa28W7RMcLiUhV7e0fEpI34fgRtBFycxIQwFdL8ysej610ZBuG1LxcQeih0vuOUKKrReyb5ajcP8aH0/KLBKs2r2"];
    
    // Create a receiver port, to receive audio
    self.input = [[ABReceiverPort alloc] initWithName:@"Receiver" title:@"Receiver"];
    _input.clientFormat = [ABMultitrackAudioEngine nonInterleavedFloatStereoAudioDescription];
    
    // Watch the connected and memberOfActiveAudiobusSession properties to be notified when we connect/disconnect or the Audiobus session begins/ends
    [_audiobusController addObserver:self forKeyPath:@"connected" options:0 context:kAudiobusConnectedOrActiveMemberChanged];
    [_audiobusController addObserver:self forKeyPath:@"memberOfActiveAudiobusSession" options:0 context:kAudiobusConnectedOrActiveMemberChanged];
    
    // Direct the receiver port to receive separate streams
    _input.receiveMixedAudio = NO;
    // _input.automaticMonitoring = YES; // If we didn't want to write our own monitoring code, we could use this instead

    [_audiobusController addReceiverPort:_input];
    
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(audiobusConnectionsChanged:) name:ABConnectionsChangedNotification object:nil];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(applicationDidEnterBackground:) name:UIApplicationDidEnterBackgroundNotification object:nil];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(applicationWillEnterForeground:) name:UIApplicationWillEnterForegroundNotification object:nil];
    
    self.receiveBlockDictionary = [NSMutableDictionary dictionary];
    
    return self;
}

-(void)dealloc {
    [_audiobusController removeObserver:self forKeyPath:@"connected"];
    [_audiobusController removeObserver:self forKeyPath:@"memberOfActiveAudiobusSession"];
    [[NSNotificationCenter defaultCenter] removeObserver:self];
    [self teardown];
}

- (void)addAudioReceiveBlock:(ABAudioReceiverBlock)block forPort:(ABPort*)port withIdentifier:(id<NSCopying>)identifier {
    struct port_entry_t *table = [self entryForPort:port];
    if ( !table ) return;
    
    if ( _receiveBlockDictionary[identifier] ) return;
    
    for ( int i=0; i<kMaxReceiveBlocks; i++ ) {
        if ( !table->blockPtrs[i] ) {
            ABAudioReceiverBlock copied = [block copy];
            table->blockPtrs[i] = (__bridge void *)(copied);
            [_receiveBlockDictionary setObject:copied forKey:identifier];
            break;
        }
    }
}

-(void)removeAudioReceiveBlockWithIdentifier:(id<NSCopying>)identifier {
    ABAudioReceiverBlock block = _receiveBlockDictionary[identifier];
    if ( !block ) return;
    
    for ( int j=0; j<kMaxSources; j++ ) {
        for ( int i=0; i<kMaxReceiveBlocks; i++ ) {
            if ( _ports[j].blockPtrs[i] == (__bridge void *)(block) ) {
                _ports[j].blockPtrs[i] = nil;
                break;
            }
        }
    }
    
    dispatch_after(dispatch_time(DISPATCH_TIME_NOW, (int64_t)(0.5 * NSEC_PER_SEC)), dispatch_get_main_queue(), ^(void){
        // Remove later, so we don't mess up the render thread; the block won't be released until this happens
        [_receiveBlockDictionary removeObjectForKey:identifier];
    });
}

- (void)setup {
    
    NSError *error = nil;
    if ( ![[AVAudioSession sharedInstance] setCategory:AVAudioSessionCategoryPlayback
                                           withOptions:AVAudioSessionCategoryOptionMixWithOthers
                                                 error:&error] ) {
        NSLog(@"Couldn't set audio session category: %@", error);
    }
    
    if ( ![[AVAudioSession sharedInstance] setActive:YES error:&error] ) {
        NSLog(@"Couldn't set audio session active: %@", error);
    }
    
    [[AVAudioSession sharedInstance] setPreferredIOBufferDuration:0.005 error:NULL];
    
    // Create the audio unit
    AudioComponentDescription desc = {
        .componentType = kAudioUnitType_Output,
        .componentSubType = kAudioUnitSubType_RemoteIO,
        .componentManufacturer = kAudioUnitManufacturer_Apple,
        .componentFlags = 0,
        .componentFlagsMask = 0
    };
    
    AudioComponent component = AudioComponentFindNext(NULL, &desc);
    checkResult(AudioComponentInstanceNew(component, &_ioUnit), "AudioComponentInstanceNew");
    
    // Enable rendering while screen locked
    UInt32 framesPerSlice = 4096;
    checkResult(AudioUnitSetProperty(_ioUnit, kAudioUnitProperty_MaximumFramesPerSlice, kAudioUnitScope_Global, 0, &framesPerSlice, sizeof(framesPerSlice)),
                "AudioUnitSetProperty(kAudioUnitProperty_MaximumFramesPerSlice");
    
    // Add render notify
    checkResult(AudioUnitAddRenderNotify(_ioUnit, audioUnitRenderNotify, (__bridge void*)self), "AudioUnitAddRenderNotify");
    
    // Create and hook up the mixer unit
    desc = (AudioComponentDescription){
        .componentType = kAudioUnitType_Mixer,
        .componentSubType = kAudioUnitSubType_MultiChannelMixer,
        .componentManufacturer = kAudioUnitManufacturer_Apple,
        .componentFlags = 0,
        .componentFlagsMask = 0
    };
    
    component = AudioComponentFindNext(NULL, &desc);
    checkResult(AudioComponentInstanceNew(component, &_mixerUnit), "AudioComponentInstanceNew");
    
    // Enable rendering while screen locked
    checkResult(AudioUnitSetProperty(_mixerUnit, kAudioUnitProperty_MaximumFramesPerSlice, kAudioUnitScope_Global, 0, &framesPerSlice, sizeof(framesPerSlice)),
                "AudioUnitSetProperty(kAudioUnitProperty_MaximumFramesPerSlice");

    checkResult(AudioUnitSetProperty(_ioUnit, kAudioUnitProperty_MakeConnection, kAudioUnitScope_Input, 0, &(AudioUnitConnection) {
        .sourceAudioUnit = _mixerUnit,
        .sourceOutputNumber = 0,
        .destInputNumber = 0
    }, sizeof(AudioUnitConnection)), "AudioUnitSetProperty(kAudioUnitProperty_MakeConnection)");
    
    // Initialize the audio units
    checkResult(AudioUnitInitialize(_ioUnit), "AudioUnitInitialize");
    checkResult(AudioUnitInitialize(_mixerUnit), "AudioUnitInitialize");
    
    // Watch for session interruptions
    self.observerToken =
    [[NSNotificationCenter defaultCenter] addObserverForName:AVAudioSessionInterruptionNotification object:nil queue:nil usingBlock:^(NSNotification *notification) {
        NSInteger type = [notification.userInfo[AVAudioSessionInterruptionTypeKey] integerValue];
        if ( type == AVAudioSessionInterruptionTypeBegan ) {
            [self stop];
        } else {
            if ( ![self start] ) {
                // Work around an iOS 7 audio interruption bug
                [self teardown];
                [self setup];
                [self start];
            }
        }
    }];
}

- (void)teardown {
    if ( self.running ) {
        [self stop];
    }
    
    if ( _ioUnit ) {
        checkResult(AudioComponentInstanceDispose(_ioUnit), "AudioComponentInstanceDispose");
        _ioUnit = NULL;
    }
    
    if ( _observerToken ) {
        [[NSNotificationCenter defaultCenter] removeObserver:_observerToken];
        self.observerToken = nil;
    }
}

- (BOOL)running {
    Boolean running = false;
    UInt32 size = sizeof(running);
    if ( !checkResult(AudioUnitGetProperty(_ioUnit, kAudioOutputUnitProperty_IsRunning, kAudioUnitScope_Global, 0, &running, &size), "AudioUnitGetProperty") ) {
        return NO;
    }
    
    return running;
}

- (void)stop {
    NSLog(@"Stopping audio engine");
    if ( _ioUnit ) {
        checkResult(AudioOutputUnitStop(_ioUnit), "AudioOutputUnitStop");
        [[AVAudioSession sharedInstance] setActive:NO error:NULL];
    }
}

- (BOOL)start {
    NSLog(@"Starting audio engine");
    NSError *error = nil;
    if ( ![[AVAudioSession sharedInstance] setActive:YES error:&error] ) {
        NSLog(@"Couldn't activate audio session: %@", error);
        return NO;
    }
    return checkResult(AudioOutputUnitStart(_ioUnit), "AudioOutputUnitStart");
}

-(NSArray *)connectedPorts {
    return _input.sources;
}

#pragma mark - Events

- (void)audiobusConnectionsChanged:(NSNotification*)notification {
    // Add new sources to our C array. We'll use the array in the Core Audio input callback, so
    // we can avoid calling any Objective-C from there, which would be bad
    NSMutableArray *added = [NSMutableArray array];
    for ( ABPort *source in _input.sources ) {
        if ( ![self entryForPort:source] ) {
            struct port_entry_t *emptySlot = [self entryForPort:nil];
            if ( emptySlot ) {
                emptySlot->pendingRemoval = NO;
                emptySlot->portPtr = (__bridge void*)(source);
                [added addObject:source];
            }
        }
    }
    
    for ( ABPort *port in added ) {
        [[NSNotificationCenter defaultCenter] postNotificationName:ABMultitrackAudioEnginePortAddedNotification
                                                            object:self
                                                          userInfo:@{ ABMultitrackAudioEnginePortKey: port }];
    }
    
    // Update the mixer to reflect the new sources
    UInt32 busCount = (UInt32)_input.sources.count;
    if ( checkResult(AudioUnitSetProperty(_mixerUnit, kAudioUnitProperty_ElementCount, kAudioUnitScope_Input, 0, &busCount, sizeof(busCount)),
                     "AudioUnitSetProperty(kAudioUnitProperty_ElementCount)") ) {
        for ( UInt32 bus=0; bus<busCount; bus++ ) {
            // Set the stream format
            AudioStreamBasicDescription clientFormat = [ABMultitrackAudioEngine nonInterleavedFloatStereoAudioDescription];
            checkResult(AudioUnitSetProperty(_mixerUnit, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Input, bus, &clientFormat, sizeof(clientFormat)),
                        "kAudioUnitProperty_StreamFormat");
            
            // Set the render callback
            AURenderCallbackStruct rcbs = { .inputProc = audioUnitRenderCallback, .inputProcRefCon = (__bridge void *)(self) };
            checkResult(AudioUnitSetProperty(_mixerUnit, kAudioUnitProperty_SetRenderCallback, kAudioUnitScope_Input, bus, &rcbs, sizeof(rcbs)),
                        "kAudioUnitProperty_SetRenderCallback");
        }
    }
    
    // Prepare to remove old sources (this will be done on the Core Audio thread, so removals are thread-safe)
    NSMutableArray *removed = [NSMutableArray array];
    for ( int i=0; i<kMaxSources; i++ ) {
        if ( _ports[i].portPtr && ![_input.sources containsObject:(__bridge id)(_ports[i].portPtr)] ) {
            [removed addObject:(__bridge id)(_ports[i].portPtr)];
            _ports[i].pendingRemoval = YES;
        }
    }
    
    for ( ABPort *port in removed ) {
        [[NSNotificationCenter defaultCenter] postNotificationName:ABMultitrackAudioEnginePortRemovedNotification
                                                            object:self
                                                          userInfo:@{ ABMultitrackAudioEnginePortKey: port }];
    }
}

-(void)applicationDidEnterBackground:(NSNotification *)notification {
    if ( !_audiobusController.connected && !_audiobusController.memberOfActiveAudiobusSession ) {
        // Fade out and stop the audio engine, suspending the app, if we're not connected, and we're not part of an active Audiobus session
        [ABAudioUnitFader fadeOutAudioUnit:_ioUnit completionBlock:^{ [self stop]; }];
    }
}

-(void)applicationWillEnterForeground:(NSNotification *)notification {
    if ( !self.running || [ABAudioUnitFader transitionsRunning] ) {
        // Start the audio system and fade in if it wasn't running
        [ABAudioUnitFader fadeInAudioUnit:_ioUnit beginBlock:^{ [self start]; } completionBlock:nil];
    }
}

-(void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context {
    if ( context == kAudiobusConnectedOrActiveMemberChanged ) {
        if ( [UIApplication sharedApplication].applicationState == UIApplicationStateBackground
                && !_audiobusController.connected
                && !_audiobusController.memberOfActiveAudiobusSession ) {
            // Audiobus session is finished. Time to sleep.
            [self stop];
        }
        
        if ( _audiobusController.connected ) {
            // Make sure we're running, if we're connected
            [self start];
        }
    } else {
        [super observeValueForKeyPath:keyPath ofObject:object change:change context:context];
    }
}

#pragma mark - Rendering

static OSStatus audioUnitRenderCallback(void *inRefCon, AudioUnitRenderActionFlags *ioActionFlags, const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber, UInt32 inNumberFrames, AudioBufferList *ioData) {
    // Use an unretained reference so we don't go retaining stuff in this realtime thread
    __unsafe_unretained ABMultitrackAudioEngine *self = (__bridge ABMultitrackAudioEngine*)inRefCon;
    
    if ( !ABReceiverPortIsConnected(self->_input) ) {
        *ioActionFlags |= kAudioUnitRenderAction_OutputIsSilence;
        return noErr;
    }
    
    // Find the port entry for this bus
    struct port_entry_t *portEntry = NULL;
    for ( int i=0, index=0; i<kMaxSources; i++ ) {
        if ( self->_ports[i].portPtr ) {
            if ( index == inBusNumber ){
                portEntry = &self->_ports[i];
                break;
            }
            index++;
        }
    }
    
    if ( !portEntry ) {
        *ioActionFlags |= kAudioUnitRenderAction_OutputIsSilence;
        return noErr;
    }
    
    // Receive audio from Audiobus
    AudioTimeStamp timestamp = *inTimeStamp;
    ABReceiverPortReceive(self->_input, (__bridge ABPort*)portEntry->portPtr, ioData, inNumberFrames, &timestamp);
    
    // Pass to our receive targets
    for ( int i=0; i<kMaxReceiveBlocks; i++ ) {
        __unsafe_unretained ABAudioReceiverBlock block = (__bridge ABAudioReceiverBlock)(portEntry->blockPtrs[i]);
        if ( block ) {
            block(ioData, inNumberFrames, &timestamp);
        }
    }
    
    return noErr;
}

static OSStatus audioUnitRenderNotify(void *inRefCon, AudioUnitRenderActionFlags *ioActionFlags, const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber, UInt32 inNumberFrames, AudioBufferList *ioData) {

    __unsafe_unretained ABMultitrackAudioEngine *self = (__bridge ABMultitrackAudioEngine*)inRefCon;
    
    if ( *ioActionFlags & kAudioUnitRenderAction_PreRender ) {
        // A little housekeeping: Remove sources pending removal (which we did in the connection change handler above)
        for ( int i=0; i<kMaxSources; i++ ) {
            if ( self->_ports[i].portPtr && self->_ports[i].pendingRemoval ) {
                memset(&self->_ports[i], 0, sizeof(struct port_entry_t));
            }
        }
    }
    
    if ( *ioActionFlags & kAudioUnitRenderAction_PostRender ) {
        if ( ABReceiverPortIsConnected(self->_input) ) {
            // End the current time interval - It's important to call this when receiving separate streams
            ABReceiverPortEndReceiveTimeInterval(self->_input);
        }
    }
    
    return noErr;
}

#pragma mark - Utils

- (struct port_entry_t*)entryForPort:(ABPort*)port {
    for ( int i=0; i<kMaxSources; i++ ) {
        if ( _ports[i].portPtr == (__bridge void*)port ) {
            return &_ports[i];
        }
    }
    return NULL;
}

+ (AudioStreamBasicDescription)nonInterleavedFloatStereoAudioDescription {
    AudioStreamBasicDescription audioDescription;
    memset(&audioDescription, 0, sizeof(audioDescription));
    audioDescription.mFormatID          = kAudioFormatLinearPCM;
    audioDescription.mFormatFlags       = kAudioFormatFlagIsFloat | kAudioFormatFlagIsPacked | kAudioFormatFlagIsNonInterleaved;
    audioDescription.mChannelsPerFrame  = 2;
    audioDescription.mBytesPerPacket    = sizeof(float);
    audioDescription.mFramesPerPacket   = 1;
    audioDescription.mBytesPerFrame     = sizeof(float);
    audioDescription.mBitsPerChannel    = 8 * sizeof(float);
    audioDescription.mSampleRate        = 44100.0;
    return audioDescription;
}

@end
