//
//  TPOscilloscopeView.m
//
//  Created by Michael Tyson on 27/07/2011.
//  Copyright (c) 2012 A Tasty Pixel. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without modification,
//  are permitted provided that the following conditions are met:
//
//  - Redistributions of source code must retain the above copyright notice, this
//    list of conditions and the following disclaimer.
//
//  - Redistributions in binary form must reproduce the above copyright notice, this
//    list of conditions and the following disclaimer in the documentation and/or
//    other materials provided with the distribution.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
//  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
//  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
//  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#import "TPOscilloscopeView.h"
#import <Accelerate/Accelerate.h>
#include <libkern/OSAtomic.h>

#define kBufferLength 2048 // In frames; higher values mean oscilloscope spans more time
#define kSkipFrames 16     // Frames to skip - higher value means faster render time, but rougher display

#if CGFLOAT_IS_DOUBLE
#define SAMPLETYPE double
#define VRAMP vDSP_vrampD
#define VSMUL vDSP_vsmulD
#define VSADD vDSP_vsaddD
#else
#define SAMPLETYPE float
#define VRAMP vDSP_vramp
#define VSMUL vDSP_vsmul
#define VSADD vDSP_vsadd
#endif

@interface TPOscilloscopeLayer : CALayer
- (void)start;
- (void)stop;
- (void)addAudio:(AudioBufferList*)audio length:(UInt32)lengthInFrames;
@property (nonatomic, strong) UIColor *lineColor;
@property (nonatomic, assign) BOOL freeze;
@end

@implementation TPOscilloscopeView
@dynamic lineColor;
@dynamic freeze;

-(id)initWithFrame:(CGRect)frame {
    if ( !(self = [super initWithFrame:frame]) ) return nil;
    
    self.backgroundColor = [UIColor clearColor];
    
    return self;
}

+(Class)layerClass {
    return [TPOscilloscopeLayer class];
}

-(void)start {
    [(TPOscilloscopeLayer*)self.layer start];
}

-(void)stop {
    [(TPOscilloscopeLayer*)self.layer stop];
}

-(UIColor *)lineColor {
    return ((TPOscilloscopeLayer*)self.layer).lineColor;
}

-(void)setLineColor:(UIColor *)lineColor {
    ((TPOscilloscopeLayer*)self.layer).lineColor = lineColor;
}

-(BOOL)freeze {
    return ((TPOscilloscopeLayer*)self.layer).freeze;
}

-(void)setFreeze:(BOOL)freeze {
    ((TPOscilloscopeLayer*)self.layer).freeze = freeze;
}

-(void)addAudio:(AudioBufferList *)audio length:(UInt32)lengthInFrames {
    if ( !((TPOscilloscopeLayer*)self.layer).freeze ) {
        [(TPOscilloscopeLayer*)self.layer addAudio:audio length:lengthInFrames];
    }
}

@end


@interface TPOscilloscopeLayer () {
    id           _timer;
    SAMPLETYPE   *_buffer;
    CGPoint     *_scratchBuffer;
    int          _buffer_head;
}
@end

@implementation TPOscilloscopeLayer
@synthesize lineColor=_lineColor;
@synthesize freeze=_freeze;

-(id)init {
    if ( !(self = [super init]) ) return nil;

    _buffer = (SAMPLETYPE*)calloc(kBufferLength, sizeof(SAMPLETYPE));
    _scratchBuffer = (CGPoint*)malloc(kBufferLength * sizeof(CGPoint));
    self.contentsScale = [[UIScreen mainScreen] scale];
    self.lineColor = [UIColor grayColor];
    
    // Disable animating view refreshes
    self.actions = [NSDictionary dictionaryWithObject:[NSNull null] forKey:@"contents"];
    
    return self;
}

- (void)start {
    if ( _timer ) return;
    
    if ( NSClassFromString(@"CADisplayLink") ) {
        _timer = [CADisplayLink displayLinkWithTarget:self selector:@selector(update)];
        ((CADisplayLink*)_timer).frameInterval = 2;
        [_timer addToRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
    } else {
        _timer = [NSTimer scheduledTimerWithTimeInterval:1.0/30.0 target:self selector:@selector(update) userInfo:nil repeats:YES];
    }
}

- (void)stop {
    if ( !_timer ) return;
    [_timer invalidate];
    _timer = nil;
}

-(void)dealloc {
    [self stop];
    free(_buffer);
    free(_scratchBuffer);
}

- (void)update {
    if ( !_freeze ) [self setNeedsDisplay];
}

-(void)setFreeze:(BOOL)freeze {
    _freeze = freeze;
    [self setNeedsDisplay];
}

#pragma mark - Rendering

-(void)drawInContext:(CGContextRef)ctx {
    CGContextSetShouldAntialias(ctx, false);
    
    // Render ring buffer as path
    CGContextSetLineWidth(ctx, 2);
    CGContextSetStrokeColorWithColor(ctx, [_lineColor CGColor]);
    
    int frames = kBufferLength-1;
    int tail = (_buffer_head+1) % kBufferLength;
    SAMPLETYPE x = 0;
    SAMPLETYPE xIncrement = (self.bounds.size.width / (float)(frames-1)) * (float)(kSkipFrames+1);
    SAMPLETYPE multiplier = self.bounds.size.height / 2.0;

    // Generate samples
    SAMPLETYPE *scratchPtr = (SAMPLETYPE*)_scratchBuffer;
    while ( frames > 0 ) {
        int framesToRender = MIN(frames, kBufferLength - tail);
        int samplesToRender = framesToRender / kSkipFrames;
        
        VRAMP(&x, &xIncrement, (SAMPLETYPE*)scratchPtr, 2, samplesToRender);
        VSMUL(&_buffer[tail], kSkipFrames, &multiplier, ((SAMPLETYPE*)scratchPtr)+1, 2, samplesToRender);
        
        scratchPtr += 2 * samplesToRender;
        x += (samplesToRender-1)*xIncrement;
        tail += framesToRender;
        if ( tail == kBufferLength ) tail = 0;
        frames -= framesToRender;
    }
    
    int sampleCount = (kBufferLength-1) / kSkipFrames;
    
    // Assign midpoint
    SAMPLETYPE midpoint = self.bounds.size.height / 2.0;
    VSADD((SAMPLETYPE*)_scratchBuffer+1, 2, &midpoint, (SAMPLETYPE*)_scratchBuffer+1, 2, sampleCount);
    
    // Render lines
    CGContextBeginPath(ctx);
    CGContextAddLines(ctx, (CGPoint*)_scratchBuffer, sampleCount);
    CGContextStrokePath(ctx);
}

#pragma mark - Callback

- (void)addAudio:(AudioBufferList*)audio length:(UInt32)frames {
    // Get a pointer to the audio buffer that we can advance
    float *audioPtr = audio->mBuffers[0].mData;
    
    // Copy in contiguous segments, wrapping around if necessary
    int remainingFrames = frames;
    while ( remainingFrames > 0 ) {
        int framesToCopy = MIN(remainingFrames, kBufferLength - _buffer_head);
        
#if CGFLOAT_IS_DOUBLE
        vDSP_vspdp(audioPtr, 1, _buffer + _buffer_head, 1, framesToCopy);
#else
        memcpy(_buffer + _buffer_head, audioPtr, framesToCopy * sizeof(float));
#endif
        audioPtr += framesToCopy;
        
        int buffer_head = _buffer_head + framesToCopy;
        if ( buffer_head == kBufferLength ) buffer_head = 0;
        OSMemoryBarrier();
        _buffer_head = buffer_head;
        remainingFrames -= framesToCopy;
    }
}

@end
