//
//  ABReceiverAudioEngine.m
//  Audiobus Samples
//
//  Created by Michael Tyson on 15/12/2013.
//  Copyright (c) 2011-2014 Audiobus. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without modification,
//  are permitted provided that the following conditions are met:
//
//  - Redistributions of source code must retain the above copyright notice, this
//    list of conditions and the following disclaimer.
//
//  - Redistributions in binary form must reproduce the above copyright notice, this
//    list of conditions and the following disclaimer in the documentation and/or
//    other materials provided with the distribution.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
//  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
//  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
//  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#import "ABReceiverAudioEngine.h"
#import <libkern/OSAtomic.h>
#import "Audiobus.h"
#import <AVFoundation/AVFoundation.h>
#import <Accelerate/Accelerate.h>
#import "TPCircularBuffer+AudioBufferList.h"

#define checkResult(result,operation) (_checkResult((result),(operation),strrchr(__FILE__, '/')+1,__LINE__))
static inline BOOL _checkResult(OSStatus result, const char *operation, const char* file, int line) {
    if ( result != noErr ) {
        NSLog(@"%s:%d: %s result %d %08X %4.4s\n", file, line, operation, (int)result, (int)result, (char*)&result);
        return NO;
    }
    return YES;
}

static void * kMemberOfActiveAudiobusSessionChanged = &kMemberOfActiveAudiobusSessionChanged;
static void * kAudiobusConnectedChanged = &kAudiobusConnectedChanged;

static const UInt32 kMaxFramesPerSlice = 4096;
static const int kMaxReceiveBlocks = 10;
static const UInt32 kPlaybackBufferLengthInFrames = 8192;

@interface ABReceiverAudioEngine () {
    AudioStreamBasicDescription _audioDescription;
    AudioUnit _audioUnit;
    ABAudioReceiverBlock _receiveBlocks[kMaxReceiveBlocks];
    ExtAudioFileRef _recordingFile;
    TPCircularBuffer _playbackBuffer;
    AudioBufferList *_playbackBufferList;
    AudioBufferList *_inputBufferList;
    BOOL _priorMonitoringEnabled;
    BOOL _receiverRecentlyDisconnectedInBackground;
}
@property (strong, nonatomic, readwrite) ABAudiobusController *audiobusController;
@property (strong, nonatomic) ABReceiverPort *receiver;
@property (strong, nonatomic) ABSenderPort *sender;
@property (nonatomic, strong) NSMutableDictionary *receiveBlockDictionary;
@property (nonatomic, strong) NSArray *observerTokens;
@property (nonatomic, copy) void (^playbackCompletionBlock)();
@property (nonatomic, strong) NSThread *playbackProducerThread;
@property (nonatomic, readwrite) BOOL playing;
@property (nonatomic, readwrite) BOOL recording;
@property (nonatomic, strong, readwrite) NSURL *recordingURL;
@end

@implementation ABReceiverAudioEngine

-(id)init {
    if ( !(self = [super init]) ) return nil;
    
    [self setup];
    
    // Create an Audiobus instance
    self.audiobusController = [[ABAudiobusController alloc] initWithApiKey:@"MCoqKkFCIFJlY2VpdmVyKioqYWJyZWNlaXZlci5hdWRpb2J1czovLyoqKlthdXJnLmFidXMucmVjdi4xXQ==:lAwDZ8EvhgnVE1EwQEyC8CeLvS4K2EBlSMK4ynmSEjhvNRHzZvCrLuNPSFoYsJsK5l4iXNCPUgEFDdamDCNEvuVWKz/U9RXsVeT4xuu1udohBAYzD0f9CRX6WCrxusbY"];
    
    // Create a receiver port, to receive audio
    self.receiver = [[ABReceiverPort alloc] initWithName:@"App Input" title:@"App Input"];
    [_audiobusController addReceiverPort:_receiver];
    _receiver.receiveMixedAudio = YES;
    _receiver.clientFormat = [ABReceiverAudioEngine nonInterleavedFloatStereoAudioDescription];
    // _receiver.automaticMonitoring = YES; // If we didn't want to write our own monitoring code, we could use this instead
    
    // Create sender port, to send audio
    self.sender = [[ABSenderPort alloc] initWithName:@"App Output"
                                               title:@"App Output"
                           audioComponentDescription:(AudioComponentDescription) {
                               .componentType = kAudioUnitType_RemoteGenerator,
                               .componentSubType = 'recv',
                               .componentManufacturer = 'abus' }];
    _sender.clientFormat = [ABReceiverAudioEngine nonInterleavedFloatStereoAudioDescription];
    [_audiobusController addSenderPort:_sender];
    
    // Report that we're okay with appearing in multiple places in the connection graph at once
    _audiobusController.allowsConnectionsToSelf = YES;
    
    // Watch the connected and memberOfActiveAudiobusSession properties to be notified when we connect/disconnect or the Audiobus session begins/ends
    [_audiobusController addObserver:self forKeyPath:@"connected" options:0 context:kAudiobusConnectedChanged];
    [_audiobusController addObserver:self forKeyPath:@"memberOfActiveAudiobusSession" options:0 context:kMemberOfActiveAudiobusSessionChanged];
    
    // Watch changes to receiver port's connections
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(receiverConnectionsChanged:) name:ABReceiverPortConnectionsChangedNotification object:_receiver];
    
    _monitorVolume = 1.0;
    _monitoringEnabled = NO;

    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(applicationDidFinishLaunching:) name:UIApplicationDidFinishLaunchingNotification object:nil];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(applicationDidEnterBackground:) name:UIApplicationDidEnterBackgroundNotification object:nil];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(applicationWillEnterForeground:) name:UIApplicationWillEnterForegroundNotification object:nil];
    
    self.receiveBlockDictionary = [NSMutableDictionary dictionary];
    
    // Init the playback buffer
    size_t audioByteCount = kPlaybackBufferLengthInFrames * _audioDescription.mBytesPerFrame * _audioDescription.mChannelsPerFrame;
    size_t controlDataByteCount = (sizeof(AudioBufferList) + sizeof(AudioBuffer)) * 30;
    TPCircularBufferInit(&_playbackBuffer, (int)(audioByteCount + controlDataByteCount));
    
    // Create the scratch buffers, for mixing input and playback and transmitting silence when not playing
    _playbackBufferList = (AudioBufferList*)malloc(sizeof(AudioBufferList) + sizeof(AudioBuffer));
    _playbackBufferList->mNumberBuffers = 2;
    for ( int i=0; i<_playbackBufferList->mNumberBuffers; i++ ) {
        _playbackBufferList->mBuffers[i].mNumberChannels = 1;
        _playbackBufferList->mBuffers[i].mDataByteSize = kMaxFramesPerSlice * sizeof(float);
        _playbackBufferList->mBuffers[i].mData = malloc(_playbackBufferList->mBuffers[i].mDataByteSize);
    }
    _inputBufferList = (AudioBufferList*)malloc(sizeof(AudioBufferList) + sizeof(AudioBuffer));
    _inputBufferList->mNumberBuffers = 2;
    for ( int i=0; i<_inputBufferList->mNumberBuffers; i++ ) {
        _inputBufferList->mBuffers[i].mNumberChannels = 1;
        _inputBufferList->mBuffers[i].mDataByteSize = kMaxFramesPerSlice * sizeof(float);
        _inputBufferList->mBuffers[i].mData = malloc(_inputBufferList->mBuffers[i].mDataByteSize);
    }
    
    return self;
}

-(void)dealloc {
    [_audiobusController removeObserver:self forKeyPath:@"connected"];
    [_audiobusController removeObserver:self forKeyPath:@"memberOfActiveAudiobusSession"];
    [[NSNotificationCenter defaultCenter] removeObserver:self];
    if ( _playbackProducerThread ) {
        [_playbackProducerThread cancel];
    }
    [self teardown];
    TPCircularBufferCleanup(&_playbackBuffer);
    if ( _playbackBufferList ) {
        for ( int i=0; i<_playbackBufferList->mNumberBuffers; i++ ) {
            free(_playbackBufferList->mBuffers[i].mData);
        }
        free(_playbackBufferList);
    }
    if ( _inputBufferList ) {
        for ( int i=0; i<_inputBufferList->mNumberBuffers; i++ ) {
            free(_inputBufferList->mBuffers[i].mData);
        }
        free(_inputBufferList);
    }
}

-(void)setMonitorVolume:(float)monitorVolume {
    _monitorVolume = monitorVolume;
    [[NSUserDefaults standardUserDefaults] setDouble:_monitorVolume forKey:@"Monitor Volume"];
}

- (void)addAudioReceiveBlock:(ABAudioReceiverBlock)block withIdentifier:(id<NSCopying>)identifier {
    if ( _receiveBlockDictionary[identifier] ) return;
    
    for ( int i=0; i<kMaxReceiveBlocks; i++ ) {
        if ( !_receiveBlocks[i] ) {
            _receiveBlocks[i] = [block copy];
            [_receiveBlockDictionary setObject:_receiveBlocks[i] forKey:identifier];
            break;
        }
    }
}

-(void)removeAudioReceiveBlockWithIdentifier:(id<NSCopying>)identifier {
    ABAudioReceiverBlock block = _receiveBlockDictionary[identifier];
    if ( !block ) return;
    
    for ( int i=0; i<kMaxReceiveBlocks; i++ ) {
        if ( _receiveBlocks[i] == block ) {
            _receiveBlocks[i] = nil;
            break;
        }
    }
    
    dispatch_after(dispatch_time(DISPATCH_TIME_NOW, (int64_t)(0.5 * NSEC_PER_SEC)), dispatch_get_main_queue(), ^(void){
        // Remove later, so we don't mess up the render thread; the block won't be released until this happens
        [_receiveBlockDictionary removeObjectForKey:identifier];
    });
}

- (void)setup {
    
    _audioDescription = [ABReceiverAudioEngine nonInterleavedFloatStereoAudioDescription];
    
    NSError *error = nil;
    if ( ![[AVAudioSession sharedInstance] setCategory:AVAudioSessionCategoryPlayAndRecord
                                           withOptions:AVAudioSessionCategoryOptionMixWithOthers | AVAudioSessionCategoryOptionDefaultToSpeaker
                                                 error:&error] ) {
        NSLog(@"Couldn't set audio session category: %@", error);
    }
    
    if ( ![[AVAudioSession sharedInstance] setActive:YES error:&error] ) {
        NSLog(@"Couldn't set audio session active: %@", error);
    }
    
    [[AVAudioSession sharedInstance] setPreferredIOBufferDuration:0.005 error:NULL];
    
    // Create the audio unit
    AudioComponentDescription desc = {
        .componentType = kAudioUnitType_Output,
        .componentSubType = kAudioUnitSubType_RemoteIO,
        .componentManufacturer = kAudioUnitManufacturer_Apple,
        .componentFlags = 0,
        .componentFlagsMask = 0
    };
    
    AudioComponent inputComponent = AudioComponentFindNext(NULL, &desc);
    checkResult(AudioComponentInstanceNew(inputComponent, &_audioUnit), "AudioComponentInstanceNew");
    
    // Set the stream format
    checkResult(AudioUnitSetProperty(_audioUnit, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Input, 0, &_audioDescription, sizeof(_audioDescription)),
                "kAudioUnitProperty_StreamFormat");
    
    // Enable input
    UInt32 enableFlag = 1;
    checkResult(AudioUnitSetProperty(_audioUnit, kAudioOutputUnitProperty_EnableIO, kAudioUnitScope_Input, 1, &enableFlag, sizeof(enableFlag)),
                "kAudioOutputUnitProperty_EnableIO");
    
    checkResult(AudioUnitSetProperty(_audioUnit, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Output, 1, &_audioDescription, sizeof(_audioDescription)),
                "kAudioUnitProperty_StreamFormat");
    
    // Set the render callback
    AURenderCallbackStruct rcbs = { .inputProc = audioUnitRenderCallback, .inputProcRefCon = (__bridge void *)(self) };
    checkResult(AudioUnitSetProperty(_audioUnit, kAudioUnitProperty_SetRenderCallback, kAudioUnitScope_Global, 0, &rcbs, sizeof(rcbs)),
                "kAudioUnitProperty_SetRenderCallback");
    
    UInt32 framesPerSlice = kMaxFramesPerSlice;
    checkResult(AudioUnitSetProperty(_audioUnit, kAudioUnitProperty_MaximumFramesPerSlice, kAudioUnitScope_Global, 0, &framesPerSlice, sizeof(framesPerSlice)),
                "AudioUnitSetProperty(kAudioUnitProperty_MaximumFramesPerSlice");
    
    [self updateChannelMap];
    
    // Initialize the audio unit
    checkResult(AudioUnitInitialize(_audioUnit), "AudioUnitInitialize");
    
    // Watch for session interruptions
    self.observerTokens = @[];
    self.observerTokens = [_observerTokens arrayByAddingObject:
    [[NSNotificationCenter defaultCenter] addObserverForName:AVAudioSessionInterruptionNotification object:nil queue:nil usingBlock:^(NSNotification *notification) {
        NSInteger type = [notification.userInfo[AVAudioSessionInterruptionTypeKey] integerValue];
        if ( type == AVAudioSessionInterruptionTypeBegan ) {
            [self stop];
        } else {
            if ( ![self start] ) {
                // Work around an iOS 7 audio interruption bug
                [self teardown];
                [self setup];
                [self start];
            }
        }
    }]];
    
    // Watch for route change notifications
    self.observerTokens = [_observerTokens arrayByAddingObject:
    [[NSNotificationCenter defaultCenter] addObserverForName:AVAudioSessionRouteChangeNotification object:nil queue:nil usingBlock:^(NSNotification *notification) {
        [self updateChannelMap];
    }]];
}

- (void)teardown {
    if ( self.running ) {
        [self stop];
    }
    
    if ( _audioUnit ) {
        checkResult(AudioComponentInstanceDispose(_audioUnit), "AudioComponentInstanceDispose");
        _audioUnit = NULL;
    }
    
    for ( id token in _observerTokens ) {
        [[NSNotificationCenter defaultCenter] removeObserver:token];
    }
    self.observerTokens = nil;
}

- (BOOL)running {
    Boolean running = false;
    UInt32 size = sizeof(running);
    if ( !checkResult(AudioUnitGetProperty(_audioUnit, kAudioOutputUnitProperty_IsRunning, kAudioUnitScope_Global, 0, &running, &size), "AudioUnitGetProperty") ) {
        return NO;
    }
    
    return running;
}

- (void)stop {
    NSLog(@"Stopping audio engine");
    if ( _audioUnit ) {
        checkResult(AudioOutputUnitStop(_audioUnit), "AudioOutputUnitStop");
        [[AVAudioSession sharedInstance] setActive:NO error:NULL];
    }
}

- (BOOL)start {
    NSLog(@"Starting audio engine");
    NSError *error = nil;
    if ( ![[AVAudioSession sharedInstance] setActive:YES error:&error] ) {
        NSLog(@"Couldn't activate audio session: %@", error);
        return NO;
    }
    return checkResult(AudioOutputUnitStart(_audioUnit), "AudioOutputUnitStart");
}

-(BOOL)playFileAtURL:(NSURL *)url completionHandler:(void (^)())block {
    if ( _playing ) {
        [self stopPlayback];
    }
    
    self.playbackCompletionBlock = block;
    
    ExtAudioFileRef playbackFile = NULL;
    
    // Open file
    if ( !checkResult(ExtAudioFileOpenURL((__bridge CFURLRef)url, &playbackFile), "ExtAudioFileOpenURL") ) {
        return NO;
    }
    
    // Apply client format
    if ( !checkResult(ExtAudioFileSetProperty(playbackFile, kExtAudioFileProperty_ClientDataFormat, sizeof(_audioDescription), &_audioDescription),
                      "ExtAudioFileSetProperty(kExtAudioFileProperty_ClientDataFormat)") ) {
        ExtAudioFileDispose(playbackFile);
        return NO;
    }
    
    self.playing = YES;
    
    // Start playback producer thread
    self.playbackProducerThread = [[NSThread alloc] initWithTarget:self selector:@selector(playbackProducerThread:) object:[NSValue valueWithPointer:playbackFile]];
    [_playbackProducerThread start];
    
    return YES;
}

-(void)stopPlayback {
    [_playbackProducerThread cancel];
    self.playbackProducerThread = nil;
    self.playing = NO;
}

-(BOOL)recordToFileAtURL:(NSURL *)url {
    if ( _recordingFile ) {
        [self stopRecordingWithCompletionHandler:nil];
    }
    
    ExtAudioFileRef recordingFile = NULL;
    
    AudioStreamBasicDescription outputDescription;
    outputDescription.mFormatFlags = kLinearPCMFormatFlagIsBigEndian | kLinearPCMFormatFlagIsSignedInteger | kLinearPCMFormatFlagIsPacked;
    outputDescription.mFormatID = kAudioFormatLinearPCM;
    outputDescription.mBitsPerChannel = 16;
    outputDescription.mChannelsPerFrame = 2;
    outputDescription.mBytesPerPacket =
    outputDescription.mBytesPerFrame = outputDescription.mChannelsPerFrame * (outputDescription.mBitsPerChannel/8);
    outputDescription.mFramesPerPacket = 1;
    outputDescription.mSampleRate = _audioDescription.mSampleRate;
    
    OSStatus result = ExtAudioFileCreateWithURL((__bridge CFURLRef)url,
                                                kAudioFileAIFFType,
                                                &outputDescription,
                                                NULL,
                                                kAudioFileFlags_EraseFile,
                                                &recordingFile);
    
    if ( !checkResult(result, "ExtAudioFileCreateWithURL") ) {
        return NO;
    }
    
    checkResult(ExtAudioFileSetProperty(recordingFile, kExtAudioFileProperty_ClientDataFormat, sizeof(_audioDescription), &_audioDescription), "ExtAudioFileSetProperty");
    checkResult(ExtAudioFileWriteAsync(recordingFile, 0, NULL), "ExtAudioFileWriteAsync");
    
    // Make sure all the above has been completed before changing the variables below
    OSMemoryBarrier();
    
    _recordingFile = recordingFile;
    self.recordingURL = url;
    self.recording = YES;
    
    return YES;
}

-(void)stopRecordingWithCompletionHandler:(void (^)(NSURL *recordingURL))block {
    ExtAudioFileRef recordingFile = _recordingFile;
    if ( !recordingFile ) return;
    
    // Make recordingFile unavailable to playback thread
    _recordingFile = NULL;
    NSURL *recordingURL = _recordingURL;
    self.recording = NO;
    self.recordingURL = nil;
    block = [block copy];
    
    // Wait a moment to make sure the audio thread isn't using recordingFile, before releasing it
    dispatch_after(dispatch_time(DISPATCH_TIME_NOW, (int64_t)(0.2 * NSEC_PER_SEC)), dispatch_get_main_queue(), ^(void){
        checkResult(ExtAudioFileDispose(recordingFile), "ExtAudioFileDispose");
        if ( block ) block(recordingURL);
        self.recordingURL = nil;
    });
}

#pragma mark - Events

-(void)applicationDidFinishLaunching:(NSNotification*)notification {
    if ( [[notification.userInfo[UIApplicationLaunchOptionsURLKey] scheme] hasSuffix:@".audiobus"] ) {
        // If this effect app has been launched from within Audiobus, we need to silence our output for a little while
        // to avoid feedback issues while the connection is established.
        if ( !_audiobusController.connected ) {
            // Mute
            self.monitoringEnabled = NO;
            
            // Set a timeout for three seconds, after which we can assume there's actually no
            // Audiobus connection forthcoming (something went wrong), and we should unmute
            dispatch_after(dispatch_time(DISPATCH_TIME_NOW, (int64_t)(3.0 * NSEC_PER_SEC)), dispatch_get_main_queue(), ^{
                [self updateMonitoringStatus];
            });
        }
    }
}

-(void)applicationDidEnterBackground:(NSNotification *)notification {
    if ( !_audiobusController.connected && !_audiobusController.memberOfActiveAudiobusSession ) {
        // Fade out and stop the audio engine, suspending the app, if we're not connected, and we're not part of an active Audiobus session
        [ABAudioUnitFader fadeOutAudioUnit:_audioUnit completionBlock:^{ [self stop]; }];
    }
}

-(void)applicationWillEnterForeground:(NSNotification *)notification {
    _receiverRecentlyDisconnectedInBackground = NO;
    if ( !self.running || [ABAudioUnitFader transitionsRunning] ) {
        // Start the audio system and fade in if it wasn't running
        [ABAudioUnitFader fadeInAudioUnit:_audioUnit beginBlock:^{ [self start]; } completionBlock:nil];
    }
}

-(void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context {
    if ( context == kMemberOfActiveAudiobusSessionChanged || context == kAudiobusConnectedChanged ) {
        if ( [UIApplication sharedApplication].applicationState == UIApplicationStateBackground
                && !_audiobusController.connected
                && !_audiobusController.memberOfActiveAudiobusSession ) {
            // Audiobus session is finished. Time to sleep.
            [self stop];
        }
        if ( context == kAudiobusConnectedChanged ) {
            if ( _audiobusController.connected ) {
                // Make sure we're running, if we're connected
                [self start];
            }
            
            // If receiver port was disconnected while we're in the background, we're going to mute
            _receiverRecentlyDisconnectedInBackground = !_receiver.connected && [UIApplication sharedApplication].applicationState == UIApplicationStateBackground;
        }
    } else {
        [super observeValueForKeyPath:keyPath ofObject:object change:change context:context];
    }
}

-(void)receiverConnectionsChanged:(NSNotification*)notification {
    [self updateMonitoringStatus];
}

-(void)updateMonitoringStatus {
    // Set monitoring status to enabled only if we're connected, to avoid the risk of feedback when disconnected (and pulling
    // audio from the mic)
    self.monitoringEnabled = _receiver.connected;
}

#pragma mark - Rendering

static OSStatus audioUnitRenderCallback(void *inRefCon, AudioUnitRenderActionFlags *ioActionFlags, const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber, UInt32 inNumberFrames, AudioBufferList *ioData) {
    // Use an unretained reference so we don't go retaining stuff in this realtime thread
    __unsafe_unretained ABReceiverAudioEngine *self = (__bridge ABReceiverAudioEngine*)inRefCon;
    
    // Reset and silence the buffers
    for ( int i=0; i<ioData->mNumberBuffers; i++ ) {
        vDSP_vclr((float*)ioData->mBuffers[i].mData, 1, inNumberFrames);
        self->_inputBufferList->mBuffers[i].mDataByteSize = kMaxFramesPerSlice * sizeof(float);
        vDSP_vclr((float*)self->_inputBufferList->mBuffers[i].mData, 1, kMaxFramesPerSlice);
        self->_playbackBufferList->mBuffers[i].mDataByteSize = kMaxFramesPerSlice * sizeof(float);
        vDSP_vclr((float*)self->_playbackBufferList->mBuffers[i].mData, 1, kMaxFramesPerSlice);
    }
    
    //
    // Input:
    //
    
    AudioTimeStamp timestamp = *inTimeStamp;
    if ( ABReceiverPortIsConnected(self->_receiver) ) {
        // Receive audio from Audiobus
        ABReceiverPortReceive(self->_receiver, nil, self->_inputBufferList, inNumberFrames, &timestamp);
    } else {
        // Receive audio from system input
        checkResult(AudioUnitRender(self->_audioUnit, ioActionFlags, inTimeStamp, 1, inNumberFrames, self->_inputBufferList), "AudioUnitRender");
    }
    
    // Pass to our receive targets
    for ( int i=0; i<kMaxReceiveBlocks; i++ ) {
        __unsafe_unretained ABAudioReceiverBlock block = self->_receiveBlocks[i];
        if ( block ) {
            block(self->_inputBufferList, inNumberFrames, &timestamp);
        }
    }
    
    //
    // Recording:
    //
    
    ExtAudioFileRef recordingFile = self->_recordingFile;
    BOOL recording = self->_recording;
    if ( recording && recordingFile ) {
        checkResult(ExtAudioFileWriteAsync(recordingFile, inNumberFrames, self->_inputBufferList), "ExtAudioFileWriteAsync");
    }

    //
    // Playback:
    //
    
    BOOL hasPlaybackAudio = NO;
    UInt32 playbackFrames = TPCircularBufferPeek(&self->_playbackBuffer, NULL, &self->_audioDescription);
    if ( playbackFrames > inNumberFrames ) playbackFrames = inNumberFrames;
    if ( playbackFrames > 0 ) {
        // Read from playback buffer
        TPCircularBufferDequeueBufferListFrames(&self->_playbackBuffer, &playbackFrames, self->_playbackBufferList, NULL, &self->_audioDescription);
        hasPlaybackAudio = YES;
    }
    
    //
    // Mixing and output:
    //

    // Apply monitor enabled/volume
    if ( !self->_monitoringEnabled ) {
        for ( int i=0; i<self->_inputBufferList->mNumberBuffers; i++ ) {
            vDSP_vclr((float*)self->_inputBufferList->mBuffers[i].mData, 1, inNumberFrames);
        }
    } else if ( self->_monitorVolume != 1.0 ) {
        for ( int i=0; i<self->_inputBufferList->mNumberBuffers; i++ ) {
            vDSP_vsmul((float*)self->_inputBufferList->mBuffers[i].mData, 1, &self->_monitorVolume, (float*)self->_inputBufferList->mBuffers[i].mData, 1, inNumberFrames);
        }
    }
    
    // Mix monitor and playback buffers into ioData
    for ( int i=0; i<ioData->mNumberBuffers; i++ ) {
        vDSP_vadd((float*)self->_inputBufferList->mBuffers[i].mData, 1,
                  (float*)self->_playbackBufferList->mBuffers[i].mData, 1,
                  (float*)ioData->mBuffers[i].mData, 1, inNumberFrames);
    }
    
    // Audiobus output
    if ( ABSenderPortIsConnectedToSelf(self->_sender) ) {
        // If we're connected to our own output, then we'll send just the playback audio, without the input mixed
        // We do this to avoid feedback (sending output based on the input, which is the output = badness)
        ABSenderPortSend(self->_sender, self->_playbackBufferList, inNumberFrames, inTimeStamp);
    } else {
        // If we're *not* connected to our own output, send the playback audio mixed with the input, as we don't need to worry about feedback.
        ABSenderPortSend(self->_sender, ioData, inNumberFrames, inTimeStamp);
    }

    // We're going to be muted if:
    if (
        //  1. The sender port is marked as muted by Audiobus - but only if we're not connected to our own receiver port (because
        //     if we are connected to ourself, we need to do monitoring or we'll go silent!).
        (ABSenderPortIsMuted(self->_sender) && !ABSenderPortIsConnectedToSelf(self->_sender)) ||
        
        //  2. Our receiver's been recently disconnected while we're in the background - this keeps the app out of the way once it's been
        //     ejected from Audiobus.
        (self->_receiverRecentlyDisconnectedInBackground && !ABSenderPortIsConnected(self->_sender)) ||
        
        //  3. We have no output, and the monitor audio is silent
        (!hasPlaybackAudio && (!self->_monitoringEnabled || self->_monitorVolume == 0.0)) ) {

        *ioActionFlags |= kAudioUnitRenderAction_OutputIsSilence;
        
    } else {
        
        // Copy buffers to output
        if ( ABSenderPortIsConnectedToSelf(self->_sender) ) {
            // If we're connected to ourself, send just the monitor audio alone, not the playback audio, which will already be in the input!
            for ( int i=0; i<ioData->mNumberBuffers; i++ ) {
                memcpy(ioData->mBuffers[i].mData, self->_inputBufferList->mBuffers[i].mData, ioData->mBuffers[i].mDataByteSize);
            }
        } else if ( !self->_monitoringEnabled ) {
            // If monitoring's disabled, send just the playback audio
            for ( int i=0; i<ioData->mNumberBuffers; i++ ) {
                memcpy(ioData->mBuffers[i].mData, self->_playbackBufferList->mBuffers[i].mData, ioData->mBuffers[i].mDataByteSize);
            }
        } else {
            // Otherwise, send the mixed playback and monitor audio already in ioData
        }
    }
    
    return noErr;
}

#pragma mark - Utils

+ (AudioStreamBasicDescription)nonInterleavedFloatStereoAudioDescription {
    AudioStreamBasicDescription audioDescription;
    memset(&audioDescription, 0, sizeof(audioDescription));
    audioDescription.mFormatID          = kAudioFormatLinearPCM;
    audioDescription.mFormatFlags       = kAudioFormatFlagIsFloat | kAudioFormatFlagIsPacked | kAudioFormatFlagIsNonInterleaved;
    audioDescription.mChannelsPerFrame  = 2;
    audioDescription.mBytesPerPacket    = sizeof(float);
    audioDescription.mFramesPerPacket   = 1;
    audioDescription.mBytesPerFrame     = sizeof(float);
    audioDescription.mBitsPerChannel    = 8 * sizeof(float);
    audioDescription.mSampleRate        = 44100.0;
    return audioDescription;
}

- (void)playbackProducerThread:(NSValue*)playbackFilePtrValue {
    NSThread *thread = [NSThread currentThread];
    ExtAudioFileRef playbackFile = (ExtAudioFileRef)[playbackFilePtrValue pointerValue];
    
    while ( !thread.isCancelled ) {
        while ( 1 ) {
            // Top up the playback buffer
            UInt32 frames = 256;
            AudioBufferList *audioBufferList = TPCircularBufferPrepareEmptyAudioBufferListWithAudioFormat(&_playbackBuffer, &_audioDescription, frames, NULL);
            if ( !audioBufferList ) break;
            
            // Read some audio
            UInt32 readFrames = frames;
            checkResult(ExtAudioFileRead(playbackFile, &readFrames, audioBufferList), "ExtAudioFileRead");
            
            if ( readFrames < frames ) {
                // We reached the end; update the audio buffer list to report the number of frames we got
                for ( int i=0; i<audioBufferList->mNumberBuffers; i++ ) {
                    audioBufferList->mBuffers[i].mDataByteSize = readFrames * _audioDescription.mBytesPerFrame;
                }
                
                // Finish playback
                dispatch_async(dispatch_get_main_queue(), ^{
                    if ( _playbackProducerThread == thread ) {
                        self.playbackProducerThread = nil;
                        self.playing = NO;
                    }
                    if ( _playbackCompletionBlock ) {
                        _playbackCompletionBlock();
                        self.playbackCompletionBlock = nil;
                    }
                });
                
                checkResult(ExtAudioFileDispose(playbackFile), "ExtAudioFileDispose");
                playbackFile = NULL;
            }
            
            // Enqueue the audio for playback
            TPCircularBufferProduceAudioBufferList(&_playbackBuffer, NULL);
            
            if ( playbackFile == NULL ) {
                break;
            }
        }
        
        if ( playbackFile == NULL ) {
            break;
        }
        
        usleep(10000);
    }
}

-(void)updateChannelMap {
    // Update channel map to ensure if we're stereo out but mono in, we hear the input in both output channels
    NSInteger numberInputChannels = [[AVAudioSession sharedInstance] inputNumberOfChannels];
    UInt32 channelMap[2] = { 0, numberInputChannels > 1 ? 1 : 0 };
    checkResult(AudioUnitSetProperty(_audioUnit, kAudioOutputUnitProperty_ChannelMap, kAudioUnitScope_Output, 1, channelMap, sizeof(channelMap)), "kAudioOutputUnitProperty_ChannelMap");
}

@end
