//
//  ABReceiverViewController.m
//  Audiobus Samples
//
//  Created by Michael Tyson on 15/12/2013.
//  Copyright (c) 2011-2014 Audiobus. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without modification,
//  are permitted provided that the following conditions are met:
//
//  - Redistributions of source code must retain the above copyright notice, this
//    list of conditions and the following disclaimer.
//
//  - Redistributions in binary form must reproduce the above copyright notice, this
//    list of conditions and the following disclaimer in the documentation and/or
//    other materials provided with the distribution.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
//  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
//  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
//  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#import "ABReceiverViewController.h"
#import "TPOscilloscopeView.h"
#import "ABReceiverAudioEngine.h"

static void * kMonitorVolumeChanged = &kMonitorVolumeChanged;
static void * kMonitorEnabledChanged = &kMonitorEnabledChanged;

@implementation ABReceiverViewController

-(id)initWithAudioEngine:(ABReceiverAudioEngine *)audioEngine {
    if ( !(self = [super initWithNibName:@"ABReceiverViewController" bundle:nil]) ) return nil;
    self.audioEngine = audioEngine;
    return self;
}

-(NSUInteger)supportedInterfaceOrientations {
    BOOL isiPad = [[UIDevice currentDevice] userInterfaceIdiom] == UIUserInterfaceIdiomPad;
    return isiPad ? UIInterfaceOrientationMaskAll : UIInterfaceOrientationMaskPortrait;
}

-(void)viewDidLoad {
    _oscilloscopeView.lineColor = [UIColor whiteColor];
}

-(void)viewWillAppear:(BOOL)animated {
    [super viewWillAppear:animated];
    
    _monitorVolumeSlider.value = _audioEngine.monitorVolume;
    _monitorEnabledButton.selected = !_audioEngine.monitoringEnabled;
    [_audioEngine addObserver:self forKeyPath:@"monitorVolume" options:0 context:kMonitorVolumeChanged];
    [_audioEngine addObserver:self forKeyPath:@"monitoringEnabled" options:0 context:kMonitorEnabledChanged];
    
    __weak TPOscilloscopeView *weakOscilloscopeView = _oscilloscopeView;
    [_audioEngine addAudioReceiveBlock:^(AudioBufferList *audio, UInt32 length, const AudioTimeStamp *timestamp) {
        TPOscilloscopeView *oscilloscopeView = weakOscilloscopeView;
        [oscilloscopeView addAudio:audio length:length];
    } withIdentifier:NSStringFromClass([self class])];
    
    [_oscilloscopeView start];
}

-(void)viewDidDisappear:(BOOL)animated {
    [super viewDidDisappear:animated];
    
    [_oscilloscopeView stop];
    
    [_audioEngine removeAudioReceiveBlockWithIdentifier:NSStringFromClass([self class])];
    [_audioEngine removeObserver:self forKeyPath:@"monitorVolume"];
    [_audioEngine removeObserver:self forKeyPath:@"monitoringEnabled"];
}

-(IBAction)toggleRecord:(id)sender {
    if ( _audioEngine.recording ) {
        [_audioEngine stopRecordingWithCompletionHandler:^(NSURL *recordingURL) {
            [[NSFileManager defaultManager] removeItemAtPath:self.recordingPath error:NULL];
            [[NSFileManager defaultManager] moveItemAtURL:recordingURL toURL:[NSURL fileURLWithPath:self.recordingPath] error:NULL];
        }];
        _recordButton.selected = NO;
    } else {
        NSString *path = self.recordingPath;
        path = [[[path stringByDeletingPathExtension] stringByAppendingString:@"-InProgress"] stringByAppendingPathExtension:[path pathExtension]];
        [_audioEngine recordToFileAtURL:[NSURL fileURLWithPath:path]];
        _recordButton.selected = YES;
    }
}

-(IBAction)togglePlay:(id)sender {
    if ( _audioEngine.playing ) {
        [_audioEngine stopPlayback];
        _playButton.selected = NO;
    } else  {
        NSString *path = self.recordingPath;
        if ( [[NSFileManager defaultManager] fileExistsAtPath:path] ) {
            BOOL success = [_audioEngine playFileAtURL:[NSURL fileURLWithPath:path] completionHandler:^{
                _playButton.selected = NO;
            }];
            
            if ( !success ) {
                NSLog(@"Couldn't play recording");
                return;
            }
            
            _playButton.selected = YES;
        }
    }
}

-(IBAction)adjustMonitorVolume:(id)sender {
    if ( !_audioEngine.monitoringEnabled ) {
        _audioEngine.monitoringEnabled = YES;
    }
    _audioEngine.monitorVolume = _monitorVolumeSlider.value;
}

-(void)toggleMonitoring:(id)sender {
    _audioEngine.monitoringEnabled = !_audioEngine.monitoringEnabled;
}

- (NSString*)recordingPath {
    NSArray * paths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
    return [[paths objectAtIndex:0] stringByAppendingPathComponent:@"Recording.aiff"];
}

-(void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context {
    if ( context == kMonitorVolumeChanged ) {
        _monitorVolumeSlider.value = ((ABReceiverAudioEngine*)object).monitorVolume;
    } else if ( context == kMonitorEnabledChanged ) {
        _monitorEnabledButton.selected = !((ABReceiverAudioEngine*)object).monitoringEnabled;
    } else {
        [super observeValueForKeyPath:keyPath ofObject:object change:change context:context];
    }
}

@end
