//
//  ABSenderAudioEngine.m
//  Audiobus Samples
//
//  Created by Michael Tyson on 17/12/2013.
//  Copyright (c) 2011-2014 Audiobus. All rights reserved.
//

#import "ABSenderAudioEngine.h"
#import <libkern/OSAtomic.h>
#import <AVFoundation/AVFoundation.h>
#import "Audiobus.h"

#define checkResult(result,operation) (_checkResult((result),(operation),strrchr(__FILE__, '/')+1,__LINE__))
static inline BOOL _checkResult(OSStatus result, const char *operation, const char* file, int line) {
    if ( result != noErr ) {
        NSLog(@"%s:%d: %s result %d %08X %4.4s\n", file, line, operation, (int)result, (int)result, (char*)&result);
        return NO;
    }
    return YES;
}

static void * kMemberOfActiveAudiobusSessionChanged = &kMemberOfActiveAudiobusSessionChanged;
static void * kAudiobusConnectedChanged = &kAudiobusConnectedChanged;
static const float kSemitoneRatio = 1.05946;
static const float kStartingFrequency = 262.0;


static const int kMaximumTones = 20;
static const int kMaxPerformBlocks = 48;

typedef enum {
    ABToneStateAttack,
    ABToneStateHold,
    ABToneStateDecay
} ABToneState;

typedef struct {
    float oscillatorFrequency;
    float oscillatorPosition;
    float gain;
    ABToneState state;
} ABTone;

typedef void (^ABPerformBlock)();

typedef struct {
    void *blockPtr;
    BOOL pending;
} ABBlockQueue;

@interface ABSenderAudioEngineProxy : NSProxy {
    ABSenderAudioEngine *_audioEngine;
}
- (id)initWithAudioEngine:(ABSenderAudioEngine*)audioController;
@end

@interface ABSenderAudioEngine () {
    AudioUnit _audioUnit;
    ABTone _tones[kMaximumTones];
    ABBlockQueue _performBlocks[kMaxPerformBlocks];
    int _attackFrames;
    int _decayFrames;
}
@property (strong, nonatomic, readwrite) ABAudiobusController *audiobusController;
@property (strong, nonatomic) ABSenderPort *sender;
@property (strong, nonatomic) NSTimer *housekeepingTimer;
@property (nonatomic, strong) id observerToken;
@end

@implementation ABSenderAudioEngine

-(id)init {
    if ( !(self = [super init]) ) return nil;
    
    [self setup];
    
    // Create an Audiobus instance
    self.audiobusController = [[ABAudiobusController alloc] initWithApiKey:@"MCoqKkFCIFNlbmRlcioqKmFic2VuZGVyLmF1ZGlvYnVzOi8vKioqW2F1cmcuYWJ1cy5zbnRoLjFd:YDV0wFnmQfuygK0r4QGcSJrdX5MwRl9AGEVJdSlX+M1IY2r4LxdB9MwcXKuQ5PkwFcDsio+qHyk3fJdsODAHhx3lRZpKCa4aVgBspiEu3LIAOEg0Cu7H3UCOO83KwoxE"];
    
    // Create a sender port
    self.sender = [[ABSenderPort alloc] initWithName:@"Synth"
                                               title:@"Synth"
                           audioComponentDescription:(AudioComponentDescription) {
                               .componentType = kAudioUnitType_RemoteGenerator,
                               .componentSubType = 'snth',
                               .componentManufacturer = 'abus' }
                                           audioUnit:_audioUnit];
    
    [_audiobusController addSenderPort:_sender];
    
    // Watch the connected and memberOfActiveAudiobusSession properties to be notified when we connect/disconnect or the Audiobus session begins/ends
    [_audiobusController addObserver:self forKeyPath:@"connected" options:0 context:kAudiobusConnectedChanged];
    [_audiobusController addObserver:self forKeyPath:@"memberOfActiveAudiobusSession" options:0 context:kMemberOfActiveAudiobusSessionChanged];
    
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(applicationDidEnterBackground:) name:UIApplicationDidEnterBackgroundNotification object:nil];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(applicationWillEnterForeground:) name:UIApplicationWillEnterForegroundNotification object:nil];
    
    self.attack = 0.002;
    self.decay = 0.4;
    self.waveform = ABSenderWaveformSine;
    
    return self;
}

-(void)dealloc {
    [_audiobusController removeObserver:self forKeyPath:@"connected"];
    [_audiobusController removeObserver:self forKeyPath:@"memberOfActiveAudiobusSession"];
    [[NSNotificationCenter defaultCenter] removeObserver:self];
    [self teardown];
}

- (void)toneOnWithFrequency:(float)frequency {
    [self queueBlockOnRenderThread:^{
        for ( int i=0; i<kMaximumTones; i++ ) {
            if ( _tones[i].oscillatorFrequency == frequency ) {
                _tones[i].state = ABToneStateAttack;
                return;
            }
        }
        for ( int i=0; i<kMaximumTones; i++ ) {
            if ( !_tones[i].oscillatorFrequency ) {
                _tones[i].state = ABToneStateAttack;
                _tones[i].gain = 0;
                _tones[i].oscillatorFrequency = frequency;
                break;
            }
        }
    }];
}

- (void)toneOffWithFrequency:(float)frequency {
    [self queueBlockOnRenderThread:^{
        for ( int i=0; i<kMaximumTones; i++ ) {
            if ( _tones[i].oscillatorFrequency == frequency ) {
                _tones[i].state = ABToneStateDecay;
                break;
            }
        }
    }];
}


- (void)toneOn:(NSUInteger) semitone{
    double frequency = kStartingFrequency * pow(kSemitoneRatio, semitone);
    [self toneOnWithFrequency:frequency];
}

- (void)toneOff:(NSUInteger)semitone{
    double frequency = kStartingFrequency * pow(kSemitoneRatio, semitone);
    [self toneOffWithFrequency:frequency];
}


- (void)allTonesOff {
    [self queueBlockOnRenderThread:^{
        for ( int i=0; i<kMaximumTones; i++ ) {
            if ( _tones[i].oscillatorFrequency ) {
                _tones[i].state = ABToneStateDecay;
            }
        }
    }];
}

-(void)setAttack:(NSTimeInterval)attack {
    if ( attack < 0.001 ) attack = 0.001;
    _attack = attack;
    _attackFrames = attack * [ABSenderAudioEngine nonInterleavedFloatStereoAudioDescription].mSampleRate;
}

-(void)setDecay:(NSTimeInterval)decay {
    _decay = decay;
    _decayFrames = decay * [ABSenderAudioEngine nonInterleavedFloatStereoAudioDescription].mSampleRate;
}

- (void)setup {
    
    NSError *error = nil;
    if ( ![[AVAudioSession sharedInstance] setCategory:AVAudioSessionCategoryPlayback
                                           withOptions:AVAudioSessionCategoryOptionMixWithOthers
                                                 error:&error] ) {
        NSLog(@"Couldn't set audio session category: %@", error);
    }
    
    if ( ![[AVAudioSession sharedInstance] setActive:YES error:&error] ) {
        NSLog(@"Couldn't set audio session active: %@", error);
    }
    
    // Create the audio unit
    AudioComponentDescription desc = {
        .componentType = kAudioUnitType_Output,
        .componentSubType = kAudioUnitSubType_RemoteIO,
        .componentManufacturer = kAudioUnitManufacturer_Apple,
        .componentFlags = 0,
        .componentFlagsMask = 0
    };
    
    AudioComponent inputComponent = AudioComponentFindNext(NULL, &desc);
    checkResult(AudioComponentInstanceNew(inputComponent, &_audioUnit), "AudioComponentInstanceNew");
    
    // Set the stream format
    AudioStreamBasicDescription clientFormat = [ABSenderAudioEngine nonInterleavedFloatStereoAudioDescription];
    checkResult(AudioUnitSetProperty(_audioUnit, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Input, 0, &clientFormat, sizeof(clientFormat)),
                "kAudioUnitProperty_StreamFormat");
    
    // Set the render callback
    AURenderCallbackStruct rcbs = { .inputProc = audioUnitRenderCallback, .inputProcRefCon = (__bridge void *)(self) };
    checkResult(AudioUnitSetProperty(_audioUnit, kAudioUnitProperty_SetRenderCallback, kAudioUnitScope_Global, 0, &rcbs, sizeof(rcbs)),
                "kAudioUnitProperty_SetRenderCallback");
    
    UInt32 framesPerSlice = 4096;
    checkResult(AudioUnitSetProperty(_audioUnit, kAudioUnitProperty_MaximumFramesPerSlice, kAudioUnitScope_Global, 0, &framesPerSlice, sizeof(framesPerSlice)),
                "AudioUnitSetProperty(kAudioUnitProperty_MaximumFramesPerSlice");
    
    // Initialize the audio unit
    checkResult(AudioUnitInitialize(_audioUnit), "AudioUnitInitialize");
    
    // Watch for session interruptions
    self.observerToken =
    [[NSNotificationCenter defaultCenter] addObserverForName:AVAudioSessionInterruptionNotification object:nil queue:nil usingBlock:^(NSNotification *notification) {
        NSInteger type = [notification.userInfo[AVAudioSessionInterruptionTypeKey] integerValue];
        if ( type == AVAudioSessionInterruptionTypeBegan ) {
            [self stop];
        } else {
            if ( ![self start] ) {
                // Work around an iOS 7 audio interruption bug
                [self teardown];
                [self setup];
                [self start];
            }
        }
    }];
}

- (void)teardown {
    if ( self.running ) {
        [self stop];
    }
    if ( _audioUnit ) {
        checkResult(AudioComponentInstanceDispose(_audioUnit), "AudioComponentInstanceDispose");
        _audioUnit = NULL;
    }
    if ( _observerToken ) {
        [[NSNotificationCenter defaultCenter] removeObserver:_observerToken];
        self.observerToken = nil;
    }
}

- (BOOL)running {
    Boolean running = false;
    UInt32 size = sizeof(running);
    if ( !checkResult(AudioUnitGetProperty(_audioUnit, kAudioOutputUnitProperty_IsRunning, kAudioUnitScope_Global, 0, &running, &size), "AudioUnitGetProperty") ) {
        return NO;
    }
    
    return running;
}

- (void)stop {
    NSLog(@"Stopping audio engine");
    [_housekeepingTimer invalidate];
    self.housekeepingTimer = nil;
    
    if ( _audioUnit ) {
        checkResult(AudioOutputUnitStop(_audioUnit), "AudioOutputUnitStop");
        [[AVAudioSession sharedInstance] setActive:NO error:NULL];
    }
}

- (BOOL)start {
    NSLog(@"Starting audio engine");
    NSError *error = nil;
    if ( ![[AVAudioSession sharedInstance] setActive:YES error:&error] ) {
        NSLog(@"Couldn't activate audio session: %@", error);
        return NO;
    }
    if ( checkResult(AudioOutputUnitStart(_audioUnit), "AudioOutputUnitStart") ) {
        if ( !_housekeepingTimer ) {
            self.housekeepingTimer = [NSTimer scheduledTimerWithTimeInterval:1.0
                                                                      target:[[ABSenderAudioEngineProxy alloc] initWithAudioEngine:self]
                                                                    selector:@selector(interThreadCleanup)
                                                                    userInfo:nil
                                                                     repeats:YES];
        }
        return YES;
    }
    return NO;
}

#pragma mark - Events

-(void)applicationDidEnterBackground:(NSNotification *)notification {
    if ( !_audiobusController.connected && !_audiobusController.memberOfActiveAudiobusSession ) {
        // Fade out and stop the audio engine, suspending the app, if we're not connected, and we're not part of an active Audiobus session
        [ABAudioUnitFader fadeOutAudioUnit:_audioUnit completionBlock:^{ [self stop]; }];
    }
}

-(void)applicationWillEnterForeground:(NSNotification *)notification {
    if ( !self.running || [ABAudioUnitFader transitionsRunning] ) {
        // Start the audio system and fade in if it wasn't running
        [ABAudioUnitFader fadeInAudioUnit:_audioUnit beginBlock:^{ [self start]; } completionBlock:nil];
    }
}

-(void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context {
    if ( context == kAudiobusConnectedChanged || context == kMemberOfActiveAudiobusSessionChanged ) {
        if ( [UIApplication sharedApplication].applicationState == UIApplicationStateBackground
                && !_audiobusController.connected
                && !_audiobusController.memberOfActiveAudiobusSession ) {
            // Audiobus session is finished. Time to sleep.
            [self stop];
        }
        if ( kAudiobusConnectedChanged && _audiobusController.connected ) {
            // Make sure we're running, if we're connected
            [self start];
        }
    } else {
        [super observeValueForKeyPath:keyPath ofObject:object change:change context:context];
    }
}

#pragma mark - Rendering

static OSStatus audioUnitRenderCallback(void *inRefCon, AudioUnitRenderActionFlags *ioActionFlags, const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber, UInt32 inNumberFrames, AudioBufferList *ioData) {
    // Use an unretained reference so we don't go retaining stuff in this realtime thread
    __unsafe_unretained ABSenderAudioEngine *self = (__bridge ABSenderAudioEngine*)inRefCon;
    
    // First run any pending blocks
    performBlocksOnRenderThread(self);
    
    // Silence buffer
    for ( int i=0; i<ioData->mNumberBuffers; i++ ) {
        memset(ioData->mBuffers[i].mData, 0, ioData->mBuffers[i].mDataByteSize);
    }
    
    // Now render tones
    int toneCount = 0;
    for ( int index=0; index<kMaximumTones; index++ ) {
        ABTone *tone = &self->_tones[index];
        if ( tone->oscillatorFrequency == 0 ) continue;
        toneCount++;
        
        float oscillatorRate = tone->oscillatorFrequency / 44100.0;
        for ( int i=0; i<inNumberFrames; i++ ) {
            float x = tone->oscillatorPosition;
            if ( self->_waveform == ABSenderWaveformSine ) {
                // Quick sin-esque oscillator
                x *= x; x -= 1.0; x *= x; x -= 0.5; x *= 0.4;
            } else if ( self->_waveform == ABSenderWaveformSaw ) {
                // Sawtooth
                x *= 0.1;
            } else if ( self->_waveform == ABSenderWaveformSquare ) {
                // Square
                x = x <= 0.0 ? -1.0 : 1.0;
                x *= 0.1;
            }
            
            tone->oscillatorPosition += oscillatorRate;
            if ( tone->oscillatorPosition > 1.0 ) tone->oscillatorPosition -= 2.0;
            
            if ( tone->state == ABToneStateAttack ) {
                x *= tone->gain;
                tone->gain += 1.0 / self->_attackFrames;
                if ( tone->gain >= 1.0 ) {
                    tone->gain = 1.0;
                    tone->state = ABToneStateHold;
                }
            } else if ( tone->state == ABToneStateDecay ) {
                x *= tone->gain;
                tone->gain -= 1.0 / self->_decayFrames;
                if ( tone->gain <= 0.0 ) {
                    memset(tone, 0, sizeof(ABTone));
                    break;
                }
            }
            
            ((float*)ioData->mBuffers[0].mData)[i] += x;
            ((float*)ioData->mBuffers[1].mData)[i] += x;
        }
    }
    
    if ( toneCount == 0 ) {
        *ioActionFlags |= kAudioUnitRenderAction_OutputIsSilence;
    }
    
    return noErr;
}

#pragma mark - Inter-thread messaging

- (void)queueBlockOnRenderThread:(ABPerformBlock)block {
    // Find insertion index to ensure order
    int nextIndex = 0;
    BOOL seenBlock = NO;
    for ( int i=0, j=0; j<kMaxPerformBlocks*2; j++, i = (i+1)%kMaxPerformBlocks ) {
        if ( _performBlocks[i].blockPtr ) {
            seenBlock = YES;
        } else if ( seenBlock ) {
            nextIndex = i;
            break;
        }
    }
    
    if ( _performBlocks[nextIndex].blockPtr ) {
        NSLog(@"No space in block buffer");
        return;
    }
    
    // Queue block
    ABPerformBlock copiedBlock = [block copy];
    _performBlocks[nextIndex].pending = YES;
    OSMemoryBarrier();
    _performBlocks[nextIndex].blockPtr = (void*)CFBridgingRetain(copiedBlock);
}

static void performBlocksOnRenderThread(ABSenderAudioEngine* self) {
    for ( int i=0; i<kMaxPerformBlocks; i++ ) {
        if ( self->_performBlocks[i].blockPtr && self->_performBlocks[i].pending ) {
            ABPerformBlock block = (__bridge ABPerformBlock)(self->_performBlocks[i].blockPtr);
            block();
            OSMemoryBarrier();
            self->_performBlocks[i].pending = NO;
        }
    }
}

- (void)interThreadCleanup {
    for ( int i=0; i<kMaxPerformBlocks; i++ ) {
        if ( _performBlocks[i].blockPtr && !_performBlocks[i].pending ) {
            void *ptr = self->_performBlocks[i].blockPtr;
            _performBlocks[i].blockPtr = NULL;
            OSMemoryBarrier();
            _performBlocks[i].pending = NO;
            CFBridgingRelease(ptr);
        }
    }
}

#pragma mark - Utils

+ (AudioStreamBasicDescription)nonInterleavedFloatStereoAudioDescription {
    AudioStreamBasicDescription audioDescription;
    memset(&audioDescription, 0, sizeof(audioDescription));
    audioDescription.mFormatID          = kAudioFormatLinearPCM;
    audioDescription.mFormatFlags       = kAudioFormatFlagIsFloat | kAudioFormatFlagIsPacked | kAudioFormatFlagIsNonInterleaved;
    audioDescription.mChannelsPerFrame  = 2;
    audioDescription.mBytesPerPacket    = sizeof(float);
    audioDescription.mFramesPerPacket   = 1;
    audioDescription.mBytesPerFrame     = sizeof(float);
    audioDescription.mBitsPerChannel    = 8 * sizeof(float);
    audioDescription.mSampleRate        = 44100.0;
    return audioDescription;
}

@end

@implementation ABSenderAudioEngineProxy
- (id)initWithAudioEngine:(ABSenderAudioEngine *)audioEngine {
    _audioEngine = audioEngine;
    return self;
}
- (NSMethodSignature *)methodSignatureForSelector:(SEL)selector {
    return [_audioEngine methodSignatureForSelector:selector];
}
- (void)forwardInvocation:(NSInvocation *)invocation {
    [invocation setTarget:_audioEngine];
    [invocation invoke];
}
@end
