//
//  ABSenderViewController.m
//  Audiobus Samples
//
//  Created by Michael Tyson on 17/12/2013.
//  Copyright (c) 2011-2014 Audiobus. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without modification,
//  are permitted provided that the following conditions are met:
//
//  - Redistributions of source code must retain the above copyright notice, this
//    list of conditions and the following disclaimer.
//
//  - Redistributions in binary form must reproduce the above copyright notice, this
//    list of conditions and the following disclaimer in the documentation and/or
//    other materials provided with the distribution.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
//  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
//  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
//  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#import "ABSenderViewController.h"
#import "ABSenderAudioEngine.h"
#import "ABKeyButton.h"
#import "ABButtonTrigger.h"
#import "ABAudiobusController.h"

static const int kSemitonesPerOctave = 12;
static const float kOctaves = 1.5;
static void * kWaveformChanged = &kWaveformChanged;
static void * kAttackChanged = &kAttackChanged;
static void * kDecayChanged = &kDecayChanged;

static float kKeyTypes[] = {
    ABKeyButtonTypeWhite, // C
    ABKeyButtonTypeBlack,
    ABKeyButtonTypeWhite,
    ABKeyButtonTypeBlack,
    ABKeyButtonTypeWhite,
    ABKeyButtonTypeWhite,
    ABKeyButtonTypeBlack,
    ABKeyButtonTypeWhite,
    ABKeyButtonTypeBlack,
    ABKeyButtonTypeWhite,
    ABKeyButtonTypeBlack,
    ABKeyButtonTypeWhite
};

@implementation ABSenderViewController

#pragma mark - View lifecycle

-(id)initWithAudioEngine:(ABSenderAudioEngine *)audioEngine {
    if ( !(self = [super initWithNibName:@"ABSenderViewController" bundle:nil]) ) return nil;
    self.audioEngine = audioEngine;
    return self;
}

-(void)dealloc {
    self.audioEngine = nil;
}

-(void)viewDidLoad {
    ABKeyButton *firstWhiteKey = nil;
    ABKeyButton *previousWhiteKey = nil;
    
    NSMutableArray *blackKeys = [NSMutableArray array];
    
    int keyCount = (float)kSemitonesPerOctave * (float)kOctaves;
    for ( int i=0; i<keyCount; i++ ) {
        ABKeyButtonType type = kKeyTypes[i%kSemitonesPerOctave];
        ABKeyButton *key = [ABKeyButton keyButtonWithType:type];
        [key addTarget:self action:@selector(keyDown:) forControlEvents:UIControlEventTouchDown];
        [key addTarget:self action:@selector(keyDown:) forControlEvents:UIControlEventTouchDragEnter];
        [key addTarget:self action:@selector(keyUp:) forControlEvents:UIControlEventTouchUpInside];
        [key addTarget:self action:@selector(keyUp:) forControlEvents:UIControlEventTouchUpOutside];
        [key addTarget:self action:@selector(keyUp:) forControlEvents:UIControlEventTouchDragExit];
        key.tag = i;
        [_keysView addSubview:key];
        
        key.translatesAutoresizingMaskIntoConstraints = NO;
        
        if ( type == ABKeyButtonTypeWhite ) {
            [_keysView addConstraint:[NSLayoutConstraint constraintWithItem:key attribute:NSLayoutAttributeTop relatedBy:NSLayoutRelationEqual toItem:_keysView attribute:NSLayoutAttributeTop multiplier:1 constant:0]];
            [_keysView addConstraint:[NSLayoutConstraint constraintWithItem:key attribute:NSLayoutAttributeBottom relatedBy:NSLayoutRelationEqual toItem:_keysView attribute:NSLayoutAttributeBottom multiplier:1 constant:0]];
            
            if ( i == 0 ) {
                [_keysView addConstraint:[NSLayoutConstraint constraintWithItem:key attribute:NSLayoutAttributeLeft relatedBy:NSLayoutRelationEqual toItem:_keysView attribute:NSLayoutAttributeLeft multiplier:1 constant:0]];
            } else if ( i == keyCount-1 ) {
                [_keysView addConstraint:[NSLayoutConstraint constraintWithItem:key attribute:NSLayoutAttributeRight relatedBy:NSLayoutRelationEqual toItem:_keysView attribute:NSLayoutAttributeRight multiplier:1 constant:0]];
            }
            
            if ( previousWhiteKey ) {
                [_keysView addConstraint:[NSLayoutConstraint constraintWithItem:key attribute:NSLayoutAttributeLeft relatedBy:NSLayoutRelationEqual toItem:previousWhiteKey attribute:NSLayoutAttributeRight multiplier:1 constant:0]];
            }
            
            if ( firstWhiteKey ) {
                [_keysView addConstraint:[NSLayoutConstraint constraintWithItem:key attribute:NSLayoutAttributeWidth relatedBy:NSLayoutRelationEqual toItem:firstWhiteKey attribute:NSLayoutAttributeWidth multiplier:1.0 constant:1]];
            } else {
                firstWhiteKey = key;
            }
            
            previousWhiteKey = key;
        } else {
            [_keysView addConstraint:[NSLayoutConstraint constraintWithItem:key attribute:NSLayoutAttributeHeight relatedBy:NSLayoutRelationEqual toItem:_keysView attribute:NSLayoutAttributeHeight multiplier:0.7 constant:1]];
            [_keysView addConstraint:[NSLayoutConstraint constraintWithItem:key attribute:NSLayoutAttributeWidth relatedBy:NSLayoutRelationEqual toItem:firstWhiteKey attribute:NSLayoutAttributeWidth multiplier:0.6 constant:0]];
            
            [_keysView addConstraint:[NSLayoutConstraint constraintWithItem:key attribute:NSLayoutAttributeTop relatedBy:NSLayoutRelationEqual toItem:_keysView attribute:NSLayoutAttributeTop multiplier:1 constant:0]];
            [_keysView addConstraint:[NSLayoutConstraint constraintWithItem:key attribute:NSLayoutAttributeCenterX relatedBy:NSLayoutRelationEqual toItem:previousWhiteKey attribute:NSLayoutAttributeRight multiplier:1 constant:0]];
            
            [blackKeys addObject:key];
        }
    }
    
    for ( ABKeyButton *key in blackKeys ) {
        [_keysView addSubview:key];
    }
    
    _waveformSelector.selectedSegmentIndex = _audioEngine.waveform;
    _attackSlider.value = _audioEngine.attack;
    _decaySlider.value = _audioEngine.decay;
    
    self.view.backgroundColor = [UIColor colorWithRed:0.294 green:0.561 blue:0.651 alpha:1.000];
}

-(void)setAudioEngine:(ABSenderAudioEngine *)audioEngine {
    if ( _audioEngine ) {
        [_audioEngine removeObserver:self forKeyPath:@"waveform"];
        [_audioEngine removeObserver:self forKeyPath:@"attack"];
        [_audioEngine removeObserver:self forKeyPath:@"decay"];
    }
    
    _audioEngine = audioEngine;
    
    if ( _audioEngine ) {
        [_audioEngine addObserver:self forKeyPath:@"waveform" options:0 context:kWaveformChanged];
        [_audioEngine addObserver:self forKeyPath:@"attack" options:0 context:kAttackChanged];
        [_audioEngine addObserver:self forKeyPath:@"decay" options:0 context:kDecayChanged];
    }
}

- (void)keyDown:(UIButton*)sender {
    
    if ( _holding ) {
        if ( sender.selected ) {
            sender.selected = NO;
            [_audioEngine toneOff:sender.tag];
        } else {
            sender.selected = YES;
            [_audioEngine toneOn:sender.tag];
        }
    } else {
        [_audioEngine toneOn:sender.tag];
    }
}



- (void)keyUp:(UIButton*)sender {
    if ( !_holding ) {
        [_audioEngine toneOff:sender.tag];
    }
}

-(IBAction)toggleHold:(UISwitch*)sender {
    self.holding = sender.isOn;
    
    if ( !_holding ) {
        for ( ABKeyButton *key in _keysView.subviews ) {
            key.selected = NO;
        }
    } else {
        for ( ABKeyButton *key in _keysView.subviews ) {
            key.selected = key.highlighted;
        }
    }
}

-(IBAction)switchWaveform:(UISegmentedControl*)sender {
    _audioEngine.waveform = (ABSenderWaveform)sender.selectedSegmentIndex;
}

-(IBAction)setAttack:(UISlider*)sender {
    _audioEngine.attack = sender.value;
}

-(IBAction)setDecay:(UISlider*)sender {
    _audioEngine.decay = sender.value;
}

-(void)setHolding:(BOOL)holding {
    if ( holding == _holding ) return;
    _holding = holding;
    
    if ( !_holding ) {
        [_audioEngine allTonesOff];
    }
    
    _holdSwitch.on = _holding;
}



-(void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context {
    if ( context == kWaveformChanged ) {
        _waveformSelector.selectedSegmentIndex = _audioEngine.waveform;
        [[NSUserDefaults standardUserDefaults] setInteger:_audioEngine.waveform forKey:@"Waveform"];
    } else if ( context == kAttackChanged ) {
        _attackSlider.value = _audioEngine.attack;
        [[NSUserDefaults standardUserDefaults] setDouble:_audioEngine.attack forKey:@"Attack"];
    } else if ( context == kDecayChanged ) {
        _decaySlider.value = _audioEngine.decay;
        [[NSUserDefaults standardUserDefaults] setDouble:_audioEngine.decay forKey:@"Decay"];
    } else {
        [super observeValueForKeyPath:keyPath ofObject:object change:change context:context];
    }
}

@end
